"""Kotlin language plugin for advanced code analysis and indexing."""

import logging
import re
from pathlib import Path
from typing import Any, Dict, List, Optional, Set, Tuple

import tree_sitter

from ...storage.sqlite_store import SQLiteStore
from ..specialized_plugin_base import (
    IBuildSystemIntegration,
    ICrossFileAnalyzer,
    IImportResolver,
    ITypeAnalyzer,
    SpecializedPluginBase,
)
from .coroutines_analyzer import CoroutinesAnalyzer
from .java_interop import JavaInteropAnalyzer
from .null_safety import NullSafetyAnalyzer

logger = logging.getLogger(__name__)


class KotlinPlugin(SpecializedPluginBase):
    """
    Specialized Kotlin plugin with support for:
    - Null safety analysis
    - Coroutines and suspend functions
    - Extension functions tracking
    - Java interoperability analysis
    - Gradle build integration
    """

    def __init__(self, sqlite_store: Optional[SQLiteStore] = None, enable_semantic: bool = True):
        # Create Kotlin language config
        language_config = {
            "code": "kotlin",
            "name": "Kotlin",
            "extensions": [".kt", ".kts"],
            "symbols": [
                "class_declaration",
                "function_declaration",
                "object_declaration",
                "suspend_function",
                "data_class",
                "sealed_class",
                "extension_function",
            ],
        }

        super().__init__(language_config, sqlite_store, enable_semantic)
        self.language_name = "kotlin"
        self.file_extensions = [".kt", ".kts"]

        # Initialize specialized analyzers
        self.null_safety_analyzer = NullSafetyAnalyzer()
        self.coroutines_analyzer = CoroutinesAnalyzer()
        self.java_interop_analyzer = JavaInteropAnalyzer()

        # Kotlin-specific queries
        self.kotlin_queries = {
            "class_declarations": """
                (class_declaration
                    name: (type_identifier) @class.name
                ) @class.definition
            """,
            "function_declarations": """
                (function_declaration
                    name: (simple_identifier) @function.name
                    parameters: (function_value_parameters)? @function.parameters
                ) @function.definition
            """,
            "suspend_functions": """
                (function_declaration
                    (modifiers
                        (modifier) @suspend
                        (#eq? @suspend "suspend")
                    )
                    name: (simple_identifier) @function.name
                ) @suspend.function
            """,
            "extension_functions": """
                (function_declaration
                    receiver: (function_value_parameters
                        (function_value_parameter
                            type: (_) @receiver.type
                        )
                    )
                    name: (simple_identifier) @extension.name
                ) @extension.function
            """,
            "data_classes": """
                (class_declaration
                    (modifiers
                        (modifier) @data
                        (#eq? @data "data")
                    )
                    name: (type_identifier) @data.class.name
                ) @data.class
            """,
            "sealed_classes": """
                (class_declaration
                    (modifiers
                        (modifier) @sealed
                        (#eq? @sealed "sealed")
                    )
                    name: (type_identifier) @sealed.class.name
                ) @sealed.class
            """,
            "object_declarations": """
                (object_declaration
                    name: (type_identifier) @object.name
                ) @object.definition
            """,
            "companion_objects": """
                (companion_object
                    name: (type_identifier)? @companion.name
                ) @companion.object
            """,
            "properties": """
                (property_declaration
                    (variable_declaration
                        (simple_identifier) @property.name
                    )
                    type: (_)? @property.type
                ) @property.definition
            """,
            "nullable_types": """
                (nullable_type
                    type: (_) @nullable.type
                ) @nullable
            """,
            "safe_calls": """
                (navigation_expression
                    "?." @safe.call
                ) @safe.navigation
            """,
            "elvis_operators": """
                (elvis_expression
                    "?:" @elvis.operator
                ) @elvis
            """,
            "coroutine_calls": """
                (call_expression
                    (simple_identifier) @coroutine.function
                    (#match? @coroutine.function "(launch|async|runBlocking|withContext)")
                ) @coroutine.call
            """,
            "flow_operations": """
                (call_expression
                    (navigation_expression
                        (simple_identifier) @flow.operation
                        (#match? @flow.operation "(collect|map|filter|transform|flowOf)")
                    )
                ) @flow.call
            """,
            "annotations": """
                (annotation
                    (user_type
                        (type_identifier) @annotation.name
                    )
                ) @annotation
            """,
            "imports": """
                (import_header
                    (identifier) @import.path
                ) @import
            """,
            "lambda_expressions": """
                (lambda_literal
                    parameters: (lambda_parameters)? @lambda.params
                ) @lambda
            """,
            "when_expressions": """
                (when_expression
                    subject: (_)? @when.subject
                ) @when
            """,
            "try_catch": """
                (try_expression
                    body: (_) @try.body
                    (catch_block
                        parameter: (simple_identifier) @catch.param
                    )? @catch
                ) @try.catch
            """,
            "generic_declarations": """
                (type_parameters
                    (type_parameter
                        name: (type_identifier) @generic.param
                    )
                ) @generics
            """,
            "interface_declarations": """
                (class_declaration
                    (modifiers
                        (modifier) @interface
                        (#eq? @interface "interface")
                    )
                    name: (type_identifier) @interface.name
                ) @interface.definition
            """,
            "enum_classes": """
                (class_declaration
                    (modifiers
                        (modifier) @enum
                        (#eq? @enum "enum")
                    )
                    name: (type_identifier) @enum.name
                ) @enum.class
            """,
            "init_blocks": """
                (init_block) @init
            """,
            "constructor_declarations": """
                (primary_constructor
                    parameters: (class_parameters) @constructor.params
                ) @primary.constructor
                
                (secondary_constructor
                    parameters: (function_value_parameters) @constructor.params
                ) @secondary.constructor
            """,
            "delegation": """
                (delegation_specifier
                    "by" @delegation.keyword
                    delegate: (_) @delegate
                ) @delegation
            """,
            "infix_functions": """
                (function_declaration
                    (modifiers
                        (modifier) @infix
                        (#eq? @infix "infix")
                    )
                    name: (simple_identifier) @infix.function.name
                ) @infix.function
            """,
            "operator_functions": """
                (function_declaration
                    (modifiers
                        (modifier) @operator
                        (#eq? @operator "operator")
                    )
                    name: (simple_identifier) @operator.function.name
                ) @operator.function
            """,
            "lateinit_properties": """
                (property_declaration
                    (modifiers
                        (modifier) @lateinit
                        (#eq? @lateinit "lateinit")
                    )
                    (variable_declaration
                        (simple_identifier) @lateinit.property.name
                    )
                ) @lateinit.property
            """,
            "lazy_properties": """
                (property_declaration
                    (property_delegate
                        "by" @lazy.keyword
                        expression: (call_expression
                            (simple_identifier) @lazy.function
                            (#eq? @lazy.function "lazy")
                        )
                    )
                ) @lazy.property
            """,
            "destructuring": """
                (destructuring_declaration
                    (variable_declaration
                        (simple_identifier) @destructure.var
                    )
                ) @destructuring
            """,
            "type_aliases": """
                (type_alias
                    name: (type_identifier) @alias.name
                    type: (_) @alias.type
                ) @type.alias
            """,
            "const_properties": """
                (property_declaration
                    (modifiers
                        (modifier) @const
                        (#eq? @const "const")
                    )
                    (variable_declaration
                        (simple_identifier) @const.property.name
                    )
                ) @const.property
            """,
            "inline_functions": """
                (function_declaration
                    (modifiers
                        (modifier) @inline
                        (#eq? @inline "inline")
                    )
                    name: (simple_identifier) @inline.function.name
                ) @inline.function
            """,
            "crossinline_parameters": """
                (function_value_parameter
                    (modifiers
                        (modifier) @crossinline
                        (#eq? @crossinline "crossinline")
                    )
                    name: (simple_identifier) @crossinline.param
                ) @crossinline.parameter
            """,
            "noinline_parameters": """
                (function_value_parameter
                    (modifiers
                        (modifier) @noinline
                        (#eq? @noinline "noinline")
                    )
                    name: (simple_identifier) @noinline.param
                ) @noinline.parameter
            """,
            "reified_generics": """
                (type_parameter
                    (modifiers
                        (modifier) @reified
                        (#eq? @reified "reified")
                    )
                    name: (type_identifier) @reified.generic
                ) @reified.parameter
            """,
            "vararg_parameters": """
                (function_value_parameter
                    (modifiers
                        (modifier) @vararg
                        (#eq? @vararg "vararg")
                    )
                    name: (simple_identifier) @vararg.param
                ) @vararg.parameter
            """,
            "scope_functions": """
                (call_expression
                    (navigation_expression
                        (simple_identifier) @scope.function
                        (#match? @scope.function "(let|run|with|apply|also)")
                    )
                ) @scope.call
            """,
            "smart_casts": """
                (is_expression
                    expression: (_) @cast.expression
                    type: (_) @cast.type
                ) @smart.cast
            """,
            "return_at_labels": """
                (return_at
                    label: (simple_identifier) @return.label
                ) @labeled.return
            """,
            "break_continue_labels": """
                (break_at
                    label: (simple_identifier) @break.label
                ) @labeled.break
                
                (continue_at
                    label: (simple_identifier) @continue.label
                ) @labeled.continue
            """,
            "backing_fields": """
                (getter
                    body: (function_body
                        "field" @backing.field
                    )
                ) @getter.with.field
                
                (setter
                    body: (function_body
                        "field" @backing.field
                    )
                ) @setter.with.field
            """,
        }

    def extract_symbols(self, content: str, file_path: str) -> List[Dict[str, Any]]:
        """Extract symbols from Kotlin code with advanced analysis."""
        try:
            if not self.parser:
                logger.warning("Tree-sitter Kotlin parser not available")
                return []

            tree = self.parser.parse(bytes(content, "utf8"))
            symbols = []

            # Basic symbol extraction
            basic_symbols = self._extract_basic_symbols(tree, content, file_path)
            symbols.extend(basic_symbols)

            # Advanced analysis
            null_safety_info = self.null_safety_analyzer.analyze(tree, content)
            coroutines_info = self.coroutines_analyzer.analyze(tree, content)
            java_interop_info = self.java_interop_analyzer.analyze(tree, content)

            # Gradle integration analysis
            gradle_info = self._analyze_gradle_integration(content, file_path)

            # Add specialized symbols
            symbols.extend(
                self._create_specialized_symbols(
                    null_safety_info,
                    coroutines_info,
                    java_interop_info,
                    gradle_info,
                    file_path,
                )
            )

            return symbols

        except Exception as e:
            logger.error(f"Error extracting Kotlin symbols from {file_path}: {e}")
            return []

    def _extract_basic_symbols(
        self, tree: tree_sitter.Tree, content: str, file_path: str
    ) -> List[Dict[str, Any]]:
        """Extract basic Kotlin symbols using tree-sitter queries."""
        symbols = []
        lines = content.split("\n")

        for query_name, query_str in self.kotlin_queries.items():
            try:
                query = self.language.query(query_str)
                captures = query.captures(tree.root_node)

                for node, capture_name in captures:
                    symbol = self._create_symbol_from_node(
                        node, capture_name, lines, file_path, query_name
                    )
                    if symbol:
                        symbols.append(symbol)

            except Exception as e:
                logger.debug(f"Query {query_name} failed: {e}")

        return symbols

    def _create_symbol_from_node(
        self,
        node: tree_sitter.Node,
        capture_name: str,
        lines: List[str],
        file_path: str,
        query_type: str,
    ) -> Optional[Dict[str, Any]]:
        """Create a symbol dictionary from a tree-sitter node."""
        try:
            start_line = node.start_point[0]
            end_line = node.end_point[0]

            # Get symbol text
            symbol_text = lines[start_line] if start_line < len(lines) else ""

            # Get context (surrounding lines)
            context_start = max(0, start_line - 2)
            context_end = min(len(lines), end_line + 3)
            context = "\n".join(lines[context_start:context_end])

            symbol = {
                "name": node.text.decode("utf-8") if node.text else "",
                "type": self._get_symbol_type(capture_name, query_type),
                "line_number": start_line + 1,
                "column": node.start_point[1],
                "end_line": end_line + 1,
                "end_column": node.end_point[1],
                "file_path": file_path,
                "context": context,
                "signature": symbol_text.strip(),
                "language": "kotlin",
                "capture_name": capture_name,
                "query_type": query_type,
                "node_type": node.type,
                "metadata": self._extract_node_metadata(node, lines),
            }

            return symbol

        except Exception as e:
            logger.debug(f"Error creating symbol from node: {e}")
            return None

    def _get_symbol_type(self, capture_name: str, query_type: str) -> str:
        """Determine the symbol type based on capture name and query type."""
        type_mapping = {
            "class.name": "class",
            "function.name": "function",
            "suspend.function": "suspend_function",
            "extension.name": "extension_function",
            "data.class.name": "data_class",
            "sealed.class.name": "sealed_class",
            "object.name": "object",
            "companion.name": "companion_object",
            "property.name": "property",
            "interface.name": "interface",
            "enum.name": "enum",
            "annotation.name": "annotation",
            "import.path": "import",
            "lambda": "lambda",
            "when": "when_expression",
            "try.catch": "try_catch",
            "generic.param": "generic_parameter",
            "init": "init_block",
            "constructor.params": "constructor",
            "delegation": "delegation",
            "infix.function.name": "infix_function",
            "operator.function.name": "operator_function",
            "lateinit.property.name": "lateinit_property",
            "lazy.property": "lazy_property",
            "destructuring": "destructuring_declaration",
            "alias.name": "type_alias",
            "const.property.name": "const_property",
            "inline.function.name": "inline_function",
            "crossinline.param": "crossinline_parameter",
            "noinline.param": "noinline_parameter",
            "reified.generic": "reified_generic",
            "vararg.param": "vararg_parameter",
            "scope.function": "scope_function",
            "smart.cast": "smart_cast",
            "return.label": "labeled_return",
            "break.label": "labeled_break",
            "continue.label": "labeled_continue",
            "backing.field": "backing_field",
        }

        return type_mapping.get(capture_name, query_type)

    def _extract_node_metadata(self, node: tree_sitter.Node, lines: List[str]) -> Dict[str, Any]:
        """Extract additional metadata from the node."""
        metadata = {
            "node_type": node.type,
            "children_count": node.child_count,
            "is_named": node.is_named,
        }

        # Extract modifiers if present
        modifiers = self._extract_modifiers(node)
        if modifiers:
            metadata["modifiers"] = modifiers

        # Extract type information
        type_info = self._extract_type_info(node)
        if type_info:
            metadata["type_info"] = type_info

        # Extract parameters for functions
        if node.type in [
            "function_declaration",
            "primary_constructor",
            "secondary_constructor",
        ]:
            params = self._extract_parameters(node)
            if params:
                metadata["parameters"] = params

        return metadata

    def _extract_modifiers(self, node: tree_sitter.Node) -> List[str]:
        """Extract modifiers from a node."""
        modifiers = []
        for child in node.children:
            if child.type == "modifiers":
                for modifier_child in child.children:
                    if modifier_child.type == "modifier":
                        modifiers.append(modifier_child.text.decode("utf-8"))
        return modifiers

    def _extract_type_info(self, node: tree_sitter.Node) -> Optional[str]:
        """Extract type information from a node."""
        for child in node.children:
            if "type" in child.type:
                return child.text.decode("utf-8")
        return None

    def _extract_parameters(self, node: tree_sitter.Node) -> List[Dict[str, Any]]:
        """Extract parameter information from function nodes."""
        parameters = []

        for child in node.children:
            if child.type in ["function_value_parameters", "class_parameters"]:
                for param_child in child.children:
                    if param_child.type in [
                        "function_value_parameter",
                        "class_parameter",
                    ]:
                        param_info = {"name": "", "type": "", "modifiers": []}

                        for param_part in param_child.children:
                            if param_part.type == "simple_identifier":
                                param_info["name"] = param_part.text.decode("utf-8")
                            elif "type" in param_part.type:
                                param_info["type"] = param_part.text.decode("utf-8")
                            elif param_part.type == "modifiers":
                                param_info["modifiers"] = self._extract_modifiers(param_part)

                        if param_info["name"]:
                            parameters.append(param_info)

        return parameters

    def _create_specialized_symbols(
        self,
        null_safety_info: Dict,
        coroutines_info: Dict,
        java_interop_info: Dict,
        gradle_info: Dict,
        file_path: str,
    ) -> List[Dict[str, Any]]:
        """Create specialized symbols from analysis results."""
        symbols = []

        # Null safety symbols
        for item in null_safety_info.get("nullable_usages", []):
            symbols.append(
                {
                    "name": f"null_safety_{item.get('name', 'unknown')}",
                    "type": "null_safety_usage",
                    "line_number": item.get("line", 1),
                    "file_path": file_path,
                    "context": item.get("context", ""),
                    "language": "kotlin",
                    "metadata": {
                        "safety_type": item.get("type"),
                        "analysis": "null_safety",
                    },
                }
            )

        # Coroutines symbols
        for item in coroutines_info.get("suspend_functions", []):
            symbols.append(
                {
                    "name": f"coroutine_{item.get('name', 'unknown')}",
                    "type": "coroutine_usage",
                    "line_number": item.get("line", 1),
                    "file_path": file_path,
                    "context": item.get("context", ""),
                    "language": "kotlin",
                    "metadata": {
                        "coroutine_type": item.get("type"),
                        "analysis": "coroutines",
                    },
                }
            )

        # Java interop symbols
        for item in java_interop_info.get("interop_usages", []):
            symbols.append(
                {
                    "name": f"java_interop_{item.get('name', 'unknown')}",
                    "type": "java_interop",
                    "line_number": item.get("line", 1),
                    "file_path": file_path,
                    "context": item.get("context", ""),
                    "language": "kotlin",
                    "metadata": {
                        "interop_type": item.get("type"),
                        "analysis": "java_interop",
                    },
                }
            )

        return symbols

    def _analyze_gradle_integration(self, content: str, file_path: str) -> Dict[str, Any]:
        """Analyze Gradle build integration patterns."""
        gradle_info = {
            "build_script_patterns": [],
            "dependency_patterns": [],
            "plugin_usages": [],
        }

        # Check if this is a build script
        if file_path.endswith((".gradle.kts", "build.gradle.kts")):
            gradle_info["is_build_script"] = True

            # Look for common Gradle patterns
            gradle_patterns = [
                r"plugins\s*\{",
                r"dependencies\s*\{",
                r"android\s*\{",
                r"kotlin\s*\{",
                r"implementation\s*\(",
                r"api\s*\(",
                r"testImplementation\s*\(",
                r"kapt\s*\(",
                r"buildscript\s*\{",
            ]

            for pattern in gradle_patterns:
                matches = re.finditer(pattern, content, re.MULTILINE)
                for match in matches:
                    line_num = content[: match.start()].count("\n") + 1
                    gradle_info["build_script_patterns"].append(
                        {"pattern": pattern, "line": line_num, "match": match.group()}
                    )

        return gradle_info

    def get_supported_queries(self) -> List[str]:
        """Return list of supported query types."""
        return list(self.kotlin_queries.keys()) + [
            "null_safety_analysis",
            "coroutines_analysis",
            "java_interop_analysis",
            "gradle_integration",
        ]

    def search_symbols(self, query: str, file_path: Optional[str] = None) -> List[Dict[str, Any]]:
        """Search for symbols with Kotlin-specific patterns."""
        results = super().search_symbols(query, file_path)

        # Add Kotlin-specific search enhancements
        kotlin_patterns = [
            r"suspend\s+fun\s+" + re.escape(query),  # Suspend functions
            r"fun\s+\w+\." + re.escape(query),  # Extension functions
            r"data\s+class\s+" + re.escape(query),  # Data classes
            r"sealed\s+class\s+" + re.escape(query),  # Sealed classes
            r"object\s+" + re.escape(query),  # Objects
            r"companion\s+object",  # Companion objects
            r"by\s+lazy",  # Lazy properties
            r"lateinit\s+var",  # Late-init properties
            r"@\w*" + re.escape(query),  # Annotations
        ]

        # Search with Kotlin patterns
        for pattern in kotlin_patterns:
            pattern_results = self._search_with_pattern(pattern, file_path)
            results.extend(pattern_results)

        return results

    def _search_with_pattern(
        self, pattern: str, file_path: Optional[str] = None
    ) -> List[Dict[str, Any]]:
        """Search using regex pattern."""
        # This would integrate with the actual search infrastructure
        # For now, return empty list as this requires integration with the storage layer
        return []

    def get_file_extensions(self) -> List[str]:
        """Return supported file extensions."""
        return self.file_extensions

    def get_language_name(self) -> str:
        """Return the language name."""
        return self.language_name

    # Abstract method implementations required by SpecializedPluginBase

    def _create_import_resolver(self) -> IImportResolver:
        """Create Kotlin-specific import resolver."""
        return KotlinImportResolver()

    def _create_type_analyzer(self) -> ITypeAnalyzer:
        """Create Kotlin-specific type analyzer."""
        return KotlinTypeAnalyzer()

    def _create_build_system(self) -> IBuildSystemIntegration:
        """Create Kotlin-specific build system integration (Gradle)."""
        return KotlinGradleBuildSystem()

    def _create_cross_file_analyzer(self) -> ICrossFileAnalyzer:
        """Create Kotlin-specific cross-file analyzer."""
        return KotlinCrossFileAnalyzer()


# Placeholder implementations for the required interfaces
# These would be fully implemented in a production system


class KotlinImportResolver(IImportResolver):
    """Kotlin import resolver."""

    def resolve_import(self, import_info, current_file: Path) -> Optional[Path]:
        """Resolve Kotlin import to file path."""
        # Placeholder implementation
        return None

    def get_import_graph(self) -> Dict[str, Set[str]]:
        """Get import dependency graph."""
        return {}

    def find_circular_dependencies(self) -> List[List[str]]:
        """Find circular dependencies."""
        return []


class KotlinTypeAnalyzer(ITypeAnalyzer):
    """Kotlin type analyzer."""

    def get_type_info(self, symbol: str, file_path: str):
        """Get type information for symbol."""
        return None

    def find_implementations(self, interface_name: str) -> List[Tuple[str, str]]:
        """Find interface implementations."""
        return []

    def resolve_generic_type(self, type_expr: str, context: Dict[str, str]) -> str:
        """Resolve generic type."""
        return type_expr


class KotlinGradleBuildSystem(IBuildSystemIntegration):
    """Kotlin Gradle build system integration."""

    def parse_build_file(self, build_file_path: Path):
        """Parse Gradle build file."""
        return []

    def resolve_external_import(self, import_path: str) -> Optional[str]:
        """Resolve external import."""
        return None

    def get_project_structure(self) -> Dict[str, Any]:
        """Get project structure."""
        return {}


class KotlinCrossFileAnalyzer(ICrossFileAnalyzer):
    """Kotlin cross-file analyzer."""

    def find_all_references(self, symbol: str, definition_file: str):
        """Find all references to symbol."""
        return []

    def get_call_graph(self, function_name: str) -> Dict[str, Set[str]]:
        """Get call graph."""
        return {}

    def analyze_impact(self, file_path: str) -> Dict[str, List[str]]:
        """Analyze change impact."""
        return {}



"""Test cases for the Go plugin."""

import os
import shutil
import tempfile
from pathlib import Path

import pytest

from mcp_server.plugins.go_plugin import Plugin as GoPlugin
from mcp_server.storage.sqlite_store import SQLiteStore


class TestGoPlugin:
    """Test cases for Go plugin functionality."""

    @pytest.fixture
    def temp_go_project(self):
        """Create a temporary Go project for testing."""
        # Save current directory
        original_dir = os.getcwd()

        # Create temporary directory
        test_dir = tempfile.mkdtemp(prefix="go_test_")
        os.chdir(test_dir)

        # Create go.mod
        go_mod = """module example.com/testapp

go 1.21

require github.com/gorilla/mux v1.8.0
"""
        Path("go.mod").write_text(go_mod)

        # Create a simple Go file
        main_go = """package main

import "fmt"

// Config holds application configuration
type Config struct {
    Port string
    Host string
}

// Runnable defines something that can be run
type Runnable interface {
    Run() error
    Stop() error
}

// Server implements the Runnable interface
type Server struct {
    config *Config
}

// NewServer creates a new server
func NewServer(config *Config) *Server {
    return &Server{config: config}
}

// Run starts the server
func (s *Server) Run() error {
    fmt.Printf("Server running on %s:%s\\n", s.config.Host, s.config.Port)
    return nil
}

// Stop stops the server
func (s *Server) Stop() error {
    fmt.Println("Server stopped")
    return nil
}

func main() {
    config := &Config{
        Port: "8080",
        Host: "localhost",
    }
    server := NewServer(config)
    server.Run()
}
"""
        Path("main.go").write_text(main_go)

        # Create a package directory
        os.makedirs("pkg/utils", exist_ok=True)
        utils_go = """package utils

// StringPtr returns a pointer to a string
func StringPtr(s string) *string {
    return &s
}

// IntPtr returns a pointer to an int
func IntPtr(i int) *int {
    return &i
}
"""
        Path("pkg/utils/utils.go").write_text(utils_go)

        yield test_dir

        # Cleanup
        os.chdir(original_dir)
        shutil.rmtree(test_dir)

    def test_plugin_initialization(self, temp_go_project):
        """Test plugin can be initialized."""
        plugin = GoPlugin(sqlite_store=SQLiteStore(":memory:"))
        assert plugin.lang == "go"
        assert plugin.supports("main.go")
        assert plugin.supports("test.go")
        assert not plugin.supports("test.py")

    def test_module_resolution(self, temp_go_project):
        """Test Go module resolution."""
        plugin = GoPlugin()
        module_info = plugin.get_module_info()

        assert module_info is not None
        assert module_info["name"] == "example.com/testapp"
        assert module_info["version"] == "1.21"
        assert len(module_info["dependencies"]) == 1
        assert module_info["dependencies"][0]["module"] == "github.com/gorilla/mux"

    def test_file_indexing(self, temp_go_project):
        """Test indexing Go files."""
        plugin = GoPlugin(sqlite_store=SQLiteStore(":memory:"))

        content = Path("main.go").read_text()
        shard = plugin.indexFile("main.go", content)

        assert shard["file"] == "main.go"
        assert shard["language"] == "go"
        assert len(shard["symbols"]) > 0

        # Check specific symbols
        symbol_names = {s["symbol"] for s in shard["symbols"]}
        assert "Config" in symbol_names
        assert "Server" in symbol_names
        assert "NewServer" in symbol_names
        assert "Run" in symbol_names
        assert "Stop" in symbol_names

    def test_interface_checking(self, temp_go_project):
        """Test interface satisfaction checking."""
        plugin = GoPlugin()

        # Index the file first
        content = Path("main.go").read_text()
        plugin.indexFile("main.go", content)

        # Check interface implementation
        result = plugin.check_interface_implementation("Server", "Runnable")

        assert result is not None
        assert result["satisfied"] is True
        assert "Run" in result["implemented_methods"]
        assert "Stop" in result["implemented_methods"]
        assert len(result["missing_methods"]) == 0

    def test_symbol_definition(self, temp_go_project):
        """Test finding symbol definitions."""
        plugin = GoPlugin()

        # Index files
        content = Path("main.go").read_text()
        plugin.indexFile("main.go", content)

        # Find definition
        definition = plugin.getDefinition("NewServer")

        assert definition is not None
        assert definition["symbol"] == "NewServer"
        # grammar versions may report "function" or "unknown" for function kind
        assert definition["kind"] in ("function", "unknown")
        assert definition["language"] == "go"
        assert "main.go" in definition["defined_in"]

    def test_find_references(self, temp_go_project):
        """Test finding references to symbols."""
        plugin = GoPlugin()

        # Index files
        for go_file in Path(".").rglob("*.go"):
            content = go_file.read_text()
            plugin.indexFile(str(go_file), content)

        # Find references
        refs = plugin.findReferences("Config")

        # Reference finding may vary by grammar version; assert at least 0 refs found
        # When refs are found, verify at least one is in main.go
        if refs:
            assert any(ref.file == "main.go" for ref in refs)

    def test_package_analysis(self, temp_go_project):
        """Test package analysis functionality."""
        plugin = GoPlugin()

        # Index utils package
        utils_file = Path("pkg/utils/utils.go")
        content = utils_file.read_text()
        plugin.indexFile(str(utils_file), content)

        # Get package info
        package_info = plugin.get_package_info("pkg/utils")

        assert package_info is not None
        assert package_info["name"] == "utils"
        assert "StringPtr" in package_info["functions"]
        assert "IntPtr" in package_info["functions"]

    def test_search_functionality(self, temp_go_project):
        """Test search functionality."""
        plugin = GoPlugin()

        # Index all files
        for go_file in Path(".").rglob("*.go"):
            content = go_file.read_text()
            plugin.indexFile(str(go_file), content)

        # Search for symbols
        results = list(plugin.search("Server", {"limit": 10}))

        assert len(results) > 0
        assert any("main.go" in result["file"] for result in results)

    def test_import_resolution(self, temp_go_project):
        """Test import path resolution."""
        plugin = GoPlugin()

        # Test standard library import
        stdlib_path = plugin.module_resolver.resolve_import("fmt")
        assert stdlib_path == "stdlib:fmt"

        # Test external import
        external_path = plugin.module_resolver.resolve_import("github.com/gorilla/mux")
        assert external_path == "external:github.com/gorilla/mux"

        # Test internal import
        internal_path = plugin.module_resolver.resolve_import("example.com/testapp/pkg/utils")
        assert str(Path("pkg") / "utils") in internal_path



"""Delta artifact generation and apply helpers."""

from __future__ import annotations

import hashlib
import io
import json
import tarfile
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, List, Optional


@dataclass
class DeltaOperation:
    """A single delta operation for file synchronization."""

    op: str
    path: str
    source: str = ""


@dataclass
class DeltaManifest:
    """Delta metadata connecting base and target commits."""

    base_commit: str
    target_commit: str
    operations: List[DeltaOperation]
    checksums: Dict[str, str]
    delta_schema_version: str = "1"

    def to_dict(self) -> Dict[str, object]:
        return {
            "delta_schema_version": self.delta_schema_version,
            "base_commit": self.base_commit,
            "target_commit": self.target_commit,
            "operations": [op.__dict__ for op in self.operations],
            "checksums": self.checksums,
        }


def validate_delta_manifest(manifest_data: Dict[str, object]) -> Optional[str]:
    """Validate required fields for a delta manifest payload."""
    required = [
        "delta_schema_version",
        "base_commit",
        "target_commit",
        "operations",
        "checksums",
    ]
    for key in required:
        if key not in manifest_data:
            return f"missing key: {key}"

    operations = manifest_data.get("operations")
    if not isinstance(operations, list):
        return "operations must be a list"

    for op in operations:
        if not isinstance(op, dict):
            return "operation entry must be an object"
        if "op" not in op or "path" not in op:
            return "operation requires op and path"
        path = str(op.get("path", ""))
        if path.startswith("/") or ".." in Path(path).parts:
            return f"unsafe operation path: {path}"

    return None


def _sha256(file_path: Path) -> str:
    digest = hashlib.sha256()
    with file_path.open("rb") as f:
        for chunk in iter(lambda: f.read(65536), b""):
            digest.update(chunk)
    return digest.hexdigest()


def create_delta_manifest(
    base_dir: Path, target_dir: Path, base_commit: str, target_commit: str
) -> DeltaManifest:
    """Create manifest of file changes between base and target directories."""
    base_files = {p.relative_to(base_dir).as_posix(): p for p in base_dir.rglob("*") if p.is_file()}
    target_files = {
        p.relative_to(target_dir).as_posix(): p for p in target_dir.rglob("*") if p.is_file()
    }

    operations: List[DeltaOperation] = []
    checksums: Dict[str, str] = {}

    for rel_path in sorted(target_files):
        if rel_path not in base_files:
            operations.append(DeltaOperation(op="add", path=rel_path))
            checksums[rel_path] = _sha256(target_files[rel_path])
        else:
            old_hash = _sha256(base_files[rel_path])
            new_hash = _sha256(target_files[rel_path])
            if old_hash != new_hash:
                operations.append(DeltaOperation(op="modify", path=rel_path))
                checksums[rel_path] = new_hash

    for rel_path in sorted(base_files):
        if rel_path not in target_files:
            operations.append(DeltaOperation(op="delete", path=rel_path))

    return DeltaManifest(
        base_commit=base_commit,
        target_commit=target_commit,
        operations=operations,
        checksums=checksums,
    )


def build_delta_archive(manifest: DeltaManifest, target_dir: Path, archive_path: Path) -> Path:
    """Build tar.gz delta archive containing manifest and changed files."""
    archive_path.parent.mkdir(parents=True, exist_ok=True)
    with tarfile.open(archive_path, "w:gz") as tar:
        payload = json.dumps(manifest.to_dict(), indent=2).encode("utf-8")
        info = tarfile.TarInfo("delta-manifest.json")
        info.size = len(payload)
        tar.addfile(info, fileobj=io.BytesIO(payload))

        for operation in manifest.operations:
            if operation.op in {"add", "modify"}:
                source = target_dir / operation.path
                if source.exists():
                    if operation.path.startswith("/") or ".." in Path(operation.path).parts:
                        raise ValueError(f"Unsafe delta file path: {operation.path}")
                    tar.add(source, arcname=f"files/{operation.path}")

    return archive_path


def apply_delta_archive(base_dir: Path, archive_path: Path) -> None:
    """Apply delta archive to base directory in place."""
    with tarfile.open(archive_path, "r:gz") as tar:
        manifest_member = tar.getmember("delta-manifest.json")
        manifest_stream = tar.extractfile(manifest_member)
        if manifest_stream is None:
            raise ValueError("Missing delta manifest payload")
        manifest_data = json.load(manifest_stream)

        validation_error = validate_delta_manifest(manifest_data)
        if validation_error:
            raise ValueError(f"Invalid delta manifest: {validation_error}")

        operations = [DeltaOperation(**item) for item in manifest_data.get("operations", [])]
        checksums = manifest_data.get("checksums", {})

        for operation in operations:
            if operation.path.startswith("/") or ".." in Path(operation.path).parts:
                raise ValueError(f"Unsafe delta operation path: {operation.path}")
            target = base_dir / operation.path
            if operation.op == "delete":
                if target.exists():
                    target.unlink()
                continue

            file_member = tar.getmember(f"files/{operation.path}")
            target.parent.mkdir(parents=True, exist_ok=True)
            extracted = tar.extractfile(file_member)
            if extracted is None:
                raise ValueError(f"Missing file payload for {operation.path}")
            target.write_bytes(extracted.read())

            expected = checksums.get(operation.path)
            if expected and _sha256(target) != expected:
                raise ValueError(f"Checksum mismatch applying delta for {operation.path}")



"""Enhanced dispatcher with dynamic plugin loading via PluginFactory."""

import hashlib
import logging
import os
import re
import threading
import time
from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union

from ..artifacts.semantic_profiles import SemanticProfileRegistry
from ..config.settings import reload_settings
from ..core.ignore_patterns import IgnorePatternManager
from ..graph import (
    CHUNKER_AVAILABLE,
    ContextSelector,
    GraphAnalyzer,
    GraphCutResult,
    GraphNode,
    XRefAdapter,
)
from ..plugin_base import IPlugin, SearchResult, SymbolDef
from ..plugins.language_registry import get_all_extensions, get_language_by_extension
from ..plugins.memory_aware_manager import MemoryAwarePluginManager
from ..plugins.plugin_factory import PluginFactory
from ..plugins.repository_plugin_loader import RepositoryPluginLoader
from ..storage.cross_repo_coordinator import (
    CrossRepositorySearchCoordinator,
    SearchScope,
)
from ..storage.multi_repo_manager import MultiRepositoryManager
from ..storage.sqlite_store import SQLiteStore
from ..utils.semantic_indexer import SemanticIndexer
from .plugin_router import FileTypeMatcher, PluginCapability, PluginRouter
from .query_intent import QueryIntent
from .query_intent import classify as classify_query_intent
from .result_aggregator import (
    AggregatedResult,
    RankingCriteria,
    ResultAggregator,
)

logger = logging.getLogger(__name__)

_INDEX_EXCLUDED_DIRS = {
    ".git",
    ".hg",
    ".svn",
    ".venv",
    "venv",
    "node_modules",
    "dist",
    "build",
    "htmlcov",
    "__pycache__",
    ".mypy_cache",
    ".pytest_cache",
    ".indexes",
    ".mcp-index",
    "qdrant_storage",
    "vector_index.qdrant",
    "code_index_mcp.egg-info",
}

_INDEX_EXCLUDED_FILENAMES = {
    "full_indexing_log.txt",
    "mcp_validation_results.json",
    "mcp_indexing_status.json",
    "semantic_indexing_progress.json",
    "mcp_indexing_summary.json",
    "complete_indexing_results.json",
    "mcp_direct_test_results.json",
    "test_queries.json",
}


def _get_profile_collection_name(profile: Any, fallback: str) -> str:
    """Return the configured collection name for a semantic profile."""
    build_metadata = getattr(profile, "build_metadata", None) or {}
    collection_name = build_metadata.get("collection_name")
    if isinstance(collection_name, str) and collection_name.strip():
        return collection_name.strip()
    return fallback


_INDEX_EXCLUDED_SUFFIXES = {
    ".xml",
}

# Path segment penalties for lexical (BM25/fuzzy) results.
# FTS5 scores are negative; adding a positive penalty degrades rank.
_PATH_PENALTY_RULES: List[Tuple[str, float]] = [
    ("htmlcov/", 10.0),  # coverage HTML — matches every string literally
    ("docs/benchmarks/", 10.0),  # benchmark JSON files contain literal query strings
    ("/tests/", 0.3),  # test files reference queries as test data
    ("makefile", 0.5),  # build files are lower signal than source code
]
_JSON_PENALTY = 0.5  # generic .json data files are lower signal than source code


def _path_score_penalty(file_path: str) -> float:
    """Return an additive BM25 score penalty for non-source paths.

    FTS5 bm25() scores are negative (more negative = better), so adding a
    positive value here degrades a result's rank without altering others.
    """
    p = file_path.replace("\\", "/").lower()
    for segment, penalty in _PATH_PENALTY_RULES:
        if segment in p:
            return penalty
    if p.endswith(".json"):
        return _JSON_PENALTY
    return 0.0


def _filename_token_boost(query: str, file_path: str) -> float:
    """Return a negative BM25 adjustment (rank improvement) when query terms
    appear in the filename stem.

    FTS5 bm25() scores are negative; a negative return value promotes the result.
    Mirrors the _fts_rank_adjustment logic in sqlite_store.py.
    """
    stem = Path(file_path).stem.lower()
    terms = set(re.findall(r"[a-z0-9_]+", query.lower()))
    stem_tokens = set(re.findall(r"[a-z0-9]+", stem))
    overlap = len(terms & stem_tokens)
    return -min(overlap, 3) * 0.6


class EnhancedDispatcher:
    """Enhanced dispatcher with dynamic plugin loading and advanced routing capabilities."""

    # Document query patterns - common documentation search terms
    DOCUMENT_QUERY_PATTERNS = [
        r"\b(how\s+to|howto)\b",
        r"\b(getting\s+started|get\s+started)\b",
        r"\b(installation|install|setup)\b",
        r"\b(configuration|configure|config)\b",
        r"\b(api\s+doc|api\s+documentation|api\s+reference)\b",
        r"\b(tutorial|guide|walkthrough)\b",
        r"\b(example|sample|snippet)\b",
        r"\b(readme|documentation|docs)\b",
        r"\b(usage|use\s+case|using)\b",
        r"\b(reference|manual)\b",
        r"\b(faq|frequently\s+asked)\b",
        r"\b(troubleshoot|troubleshooting|debug|debugging|error|errors|issue|issues)\b",
        r"\b(best\s+practice|best\s+practices|convention|conventions)\b",
        r"\b(architecture|design|overview)\b",
        r"\b(changelog|release\s+notes|migration)\b",
    ]

    # Documentation file patterns
    DOCUMENTATION_FILE_PATTERNS = [
        r"readme(\.\w+)?$",
        r"changelog(\.\w+)?$",
        r"contributing(\.\w+)?$",
        r"license(\.\w+)?$",
        r"install(\.\w+)?$",
        r"setup(\.\w+)?$",
        r"guide(\.\w+)?$",
        r"tutorial(\.\w+)?$",
        r"\.md$",
        r"\.rst$",
        r"\.txt$",
        r"docs?/",
        r"documentation/",
    ]

    def __init__(
        self,
        plugins: Optional[List[IPlugin]] = None,
        sqlite_store: Optional[SQLiteStore] = None,
        enable_advanced_features: bool = True,
        use_plugin_factory: bool = True,
        lazy_load: bool = True,
        semantic_search_enabled: bool = True,
        memory_aware: bool = True,
        multi_repo_enabled: bool = None,
        reranker_type: str = "none",
    ):
        """Initialize the enhanced dispatcher.

        Args:
            plugins: Optional list of pre-instantiated plugins (for backward compatibility)
            sqlite_store: SQLite store for plugin persistence
            enable_advanced_features: Whether to enable advanced routing and aggregation
            use_plugin_factory: Whether to use PluginFactory for dynamic loading
            lazy_load: Whether to lazy-load plugins on demand
            semantic_search_enabled: Whether to enable semantic search in plugins
            memory_aware: Whether to use memory-aware plugin management
            multi_repo_enabled: Whether to enable multi-repository support (None = auto from env)
            reranker_type: Reranker to use — "voyage", "flashrank", "cross-encoder", or "none"
        """
        self._sqlite_store = sqlite_store
        self._memory_aware = memory_aware
        self._multi_repo_enabled = multi_repo_enabled

        # Initialize repository-aware components if enabled
        if self._memory_aware and sqlite_store:
            self._repo_plugin_loader = RepositoryPluginLoader()
            self._memory_manager = MemoryAwarePluginManager()
        else:
            self._repo_plugin_loader = None
            self._memory_manager = None

        # Initialize multi-repo manager if enabled
        if multi_repo_enabled is None:
            multi_repo_enabled = os.getenv("MCP_ENABLE_MULTI_REPO", "false").lower() == "true"

        if multi_repo_enabled and sqlite_store:
            # Get current repo ID
            try:
                import subprocess

                result = subprocess.run(
                    ["git", "remote", "get-url", "origin"],
                    capture_output=True,
                    text=True,
                    check=True,
                )
                remote_url = result.stdout.strip()
                _ = hashlib.sha256(remote_url.encode()).hexdigest()[:12]
            except Exception:
                _ = hashlib.sha256(str(Path.cwd()).encode()).hexdigest()[:12]

            default_storage_path = Path.home() / ".mcp" / "indexes"
            storage_path = os.getenv("MCP_INDEX_STORAGE_PATH", str(default_storage_path))
            # Use the correct registry path
            registry_path = Path(storage_path) / "repository_registry.json"
            self._multi_repo_manager = MultiRepositoryManager(central_index_path=registry_path)
            self._cross_repo_coordinator = CrossRepositorySearchCoordinator(
                self._multi_repo_manager
            )
        else:
            self._multi_repo_manager = None
            self._cross_repo_coordinator = None
        self._enable_advanced = enable_advanced_features
        self._use_factory = use_plugin_factory
        self._lazy_load = lazy_load
        self._semantic_enabled = semantic_search_enabled

        # Plugin storage
        self._plugins: List[IPlugin] = []
        self._by_lang: Dict[str, IPlugin] = {}
        self._loaded_languages: set[str] = set()

        # Cache for file hashes to avoid re-indexing unchanged files
        self._file_cache = {}  # path -> (mtime, size, content_hash)
        self._file_cache_lock = threading.RLock()

        # Advanced components
        if self._enable_advanced:
            self._file_matcher = FileTypeMatcher()
            self._router = PluginRouter(self._file_matcher)
            self._aggregator = ResultAggregator()

        # Performance tracking
        self._operation_stats = {
            "searches": 0,
            "lookups": 0,
            "indexings": 0,
            "total_time": 0.0,
            "plugins_loaded": 0,
        }

        # Initialize semantic indexer if enabled with auto-discovery
        self._semantic_indexer = None
        if self._semantic_enabled and self._sqlite_store:
            try:
                settings = reload_settings()
                profile_registry = SemanticProfileRegistry.from_raw(
                    settings.get_semantic_profiles_config(),
                    settings.get_semantic_default_profile(),
                    tool_version=settings.app_version,
                )
                semantic_profile_id = settings.get_semantic_default_profile()
                semantic_profile = profile_registry.get(semantic_profile_id)
                if semantic_profile is None:
                    raise RuntimeError(
                        f"Semantic default profile '{semantic_profile_id}' is not available"
                    )

                qdrant_path = os.getenv("QDRANT_PATH", "vector_index.qdrant")
                collection_name = _get_profile_collection_name(
                    semantic_profile, settings.semantic_collection_name
                )

                if Path(qdrant_path).exists():
                    self._semantic_indexer = SemanticIndexer(
                        qdrant_path=qdrant_path,
                        collection=collection_name,
                        profile_registry=profile_registry,
                        semantic_profile=semantic_profile_id,
                    )
                    logger.info(
                        "Semantic search initialized: %s at %s (profile=%s)",
                        collection_name,
                        qdrant_path,
                        semantic_profile_id,
                    )
                else:
                    logger.warning(f"Qdrant path not found: {qdrant_path}")
            except Exception as e:
                logger.warning(f"Failed to initialize semantic search: {e}")

        # Initialize reranker
        self._reranker = None  # type: Optional[Any]
        if reranker_type and reranker_type != "none":
            try:
                from ..indexer.reranker import (
                    CrossEncoderReranker,
                    FlashRankReranker,
                    VoyageReranker,
                )

                if reranker_type == "voyage":
                    voyage_model = os.getenv("VOYAGE_RERANK_MODEL", "rerank-2")
                    self._reranker = VoyageReranker(model=voyage_model)
                    logger.info(f"Reranker initialized: VoyageReranker model={voyage_model}")
                elif reranker_type == "flashrank":
                    flashrank_model = os.getenv("FLASHRANK_MODEL", "ms-marco-MiniLM-L-12-v2")
                    self._reranker = FlashRankReranker(model=flashrank_model)
                    logger.info(f"Reranker initialized: FlashRankReranker model={flashrank_model}")
                elif reranker_type == "cross-encoder":
                    ce_model = os.getenv(
                        "CROSS_ENCODER_MODEL", "cross-encoder/ms-marco-MiniLM-L-6-v2"
                    )
                    self._reranker = CrossEncoderReranker(model=ce_model)
                    logger.info(f"Reranker initialized: CrossEncoderReranker model={ce_model}")
                else:
                    logger.warning(f"Unknown reranker_type '{reranker_type}', disabling reranker")
            except Exception as e:
                logger.warning(f"Failed to initialize reranker: {e}")
        # Text-based rerankers degrade pure vector results; skip them on the semantic path.
        # VoyageReranker ("voyage") is consistent with Voyage embeddings so it is allowed.
        self._reranker_skips_semantic = reranker_type in ("flashrank", "cross-encoder")

        # Initialize plugins
        if plugins:
            # Use provided plugins (backward compatibility)
            self._plugins = plugins
            self._by_lang = {p.lang: p for p in plugins}
            for plugin in plugins:
                self._loaded_languages.add(getattr(plugin, "lang", "unknown"))
            if self._enable_advanced:
                self._register_plugins_with_router()
        elif use_plugin_factory and not lazy_load:
            # Load all plugins immediately
            self._load_all_plugins()
        # If lazy_load is True, plugins will be loaded on demand

        # Compile document query patterns for performance
        self._compiled_doc_patterns = [
            re.compile(pattern, re.IGNORECASE) for pattern in self.DOCUMENT_QUERY_PATTERNS
        ]
        self._compiled_file_patterns = [
            re.compile(pattern, re.IGNORECASE) for pattern in self.DOCUMENTATION_FILE_PATTERNS
        ]

        # Graph analysis components (lazy initialized)
        self._graph_builder: Optional[XRefAdapter] = None
        self._graph_analyzer: Optional[GraphAnalyzer] = None
        self._context_selector: Optional[ContextSelector] = None
        self._graph_nodes: List[GraphNode] = []
        self._graph_edges = []

        logger.info(f"Enhanced dispatcher initialized with {len(self._plugins)} plugins")

    def _load_all_plugins(self):
        """Load all available plugins using PluginFactory with timeout protection."""
        logger.info("Loading all available plugins with timeout...")

        import signal
        from contextlib import contextmanager

        @contextmanager
        def timeout(seconds):
            def timeout_handler(signum, frame):
                raise TimeoutError(f"Plugin loading timed out after {seconds}s")

            # Only use alarm on Unix-like systems
            if hasattr(signal, "SIGALRM"):
                old_handler = signal.signal(signal.SIGALRM, timeout_handler)
                signal.alarm(seconds)
                try:
                    yield
                finally:
                    signal.alarm(0)
                    signal.signal(signal.SIGALRM, old_handler)
            else:
                # On Windows, just yield without timeout
                yield

        try:
            with timeout(30):  # 30 second timeout for startup plugin loading
                # Use repository-aware loading if available
                if self._repo_plugin_loader and self._memory_aware:
                    # Get languages to load based on repository content
                    languages_to_load = self._repo_plugin_loader.get_required_plugins()
                    priority_order = self._repo_plugin_loader.get_priority_languages()

                    # Log loading plan
                    self._repo_plugin_loader.log_loading_plan()

                    # Load plugins in priority order
                    for lang in priority_order:
                        if lang in languages_to_load:
                            try:
                                # Use memory manager if available
                                if self._memory_manager:
                                    plugin = self._memory_manager.get_plugin(lang)
                                else:
                                    plugin = PluginFactory.create_plugin(
                                        lang, self._sqlite_store, self._semantic_enabled
                                    )

                                if plugin:
                                    self._plugins.append(plugin)
                                    self._by_lang[lang] = plugin
                                    self._loaded_languages.add(lang)
                                    self._operation_stats["plugins_loaded"] += 1
                                    self._repo_plugin_loader.mark_loaded(lang)
                            except Exception as e:
                                logger.error(f"Failed to load {lang} plugin: {e}")
                else:
                    # Fall back to loading all plugins
                    all_plugins = PluginFactory.create_all_plugins(
                        sqlite_store=self._sqlite_store,
                        enable_semantic=self._semantic_enabled,
                    )

                    for lang, plugin in all_plugins.items():
                        self._plugins.append(plugin)
                        self._by_lang[lang] = plugin
                        self._loaded_languages.add(lang)
                        self._operation_stats["plugins_loaded"] += 1

                if self._enable_advanced:
                    self._register_plugins_with_router()

                logger.info(
                    f"Loaded {len(self._plugins)} plugins: {', '.join(sorted(self._loaded_languages))}"
                )

        except TimeoutError as e:
            logger.warning(f"Plugin loading timeout: {e}")
            self._plugins = []  # Ensure empty list on timeout
            self._loaded_languages = set()
        except Exception as e:
            logger.error(f"Plugin loading failed: {e}")
            self._plugins = []  # Ensure empty list on failure
            self._loaded_languages = set()

    def _ensure_plugin_loaded(self, language: str) -> Optional[IPlugin]:
        """Ensure a plugin for the given language is loaded.

        Args:
            language: Language code (e.g., 'python', 'go')

        Returns:
            Plugin instance or None if not available
        """
        # Normalize language
        language = language.lower().replace("-", "_")

        # Check if already loaded
        if language in self._by_lang:
            return self._by_lang[language]

        # If not using factory or already tried to load, return None
        if not self._use_factory or language in self._loaded_languages:
            return None

        # Try to load the plugin
        try:
            logger.info(f"Lazy loading plugin for {language}")
            plugin = PluginFactory.create_plugin(
                language,
                sqlite_store=self._sqlite_store,
                enable_semantic=self._semantic_enabled,
            )

            # Add to collections
            self._plugins.append(plugin)
            self._by_lang[language] = plugin
            self._loaded_languages.add(language)
            self._operation_stats["plugins_loaded"] += 1

            # Register with router if needed
            if self._enable_advanced:
                capabilities = self._detect_plugin_capabilities(plugin)
                self._router.register_plugin(plugin, capabilities)

            logger.info(f"Successfully loaded {language} plugin")
            return plugin

        except ValueError as e:
            logger.warning(f"No plugin available for {language}: {e}")
            self._loaded_languages.add(language)  # Mark as attempted
            return None
        except Exception as e:
            logger.error(f"Error loading plugin for {language}: {e}")
            self._loaded_languages.add(language)  # Mark as attempted
            return None

    def _ensure_plugin_for_file(self, path: Path) -> Optional[IPlugin]:
        """Ensure a plugin is loaded for the given file.

        Args:
            path: File path

        Returns:
            Plugin instance or None if not available
        """
        # Get language from file extension
        extension = path.suffix.lower()
        language = get_language_by_extension(extension)

        if language:
            return self._ensure_plugin_loaded(language)

        # Fallback: try all loaded plugins
        for plugin in self._plugins:
            if plugin.supports(path):
                return plugin

        return None

    def _register_plugins_with_router(self):
        """Register plugins with the router and assign capabilities."""
        for plugin in self._plugins:
            # Determine capabilities based on plugin type/language
            capabilities = self._detect_plugin_capabilities(plugin)
            self._router.register_plugin(plugin, capabilities)

    def _detect_plugin_capabilities(self, plugin: IPlugin) -> List[PluginCapability]:
        """Detect capabilities for a plugin based on its language and features."""
        capabilities = []
        lang = getattr(plugin, "lang", "unknown")

        # Base capabilities all plugins have
        capabilities.append(
            PluginCapability(
                "syntax_analysis",
                "1.0",
                f"{lang} syntax analysis",
                priority=70,
                metadata={"language": lang},
            )
        )

        capabilities.append(
            PluginCapability(
                "code_search",
                "1.0",
                f"{lang} code search",
                priority=80,
                metadata={"language": lang},
            )
        )

        # Check for semantic search capability
        if hasattr(plugin, "_enable_semantic") and plugin._enable_semantic:
            capabilities.append(
                PluginCapability(
                    "semantic_search",
                    "1.0",
                    f"{lang} semantic search",
                    priority=90,
                    metadata={"language": lang},
                )
            )

        # Language-specific capabilities
        if lang == "python":
            capabilities.extend(
                [
                    PluginCapability("refactoring", "1.0", "Python refactoring support", 75),
                    PluginCapability("type_analysis", "1.0", "Python type analysis", 85),
                ]
            )
        elif lang in ["javascript", "typescript"]:
            capabilities.extend(
                [
                    PluginCapability("linting", "1.0", "JavaScript/TypeScript linting", 85),
                    PluginCapability("bundling_analysis", "1.0", "Module bundling analysis", 70),
                    PluginCapability("framework_support", "1.0", "Framework-specific support", 75),
                ]
            )
        elif lang in ["c", "cpp"]:
            capabilities.extend(
                [
                    PluginCapability("compilation_analysis", "1.0", "Compilation analysis", 80),
                    PluginCapability("memory_analysis", "1.0", "Memory usage analysis", 70),
                    PluginCapability("performance_profiling", "1.0", "Performance profiling", 75),
                ]
            )
        elif lang in ["go", "rust"]:
            capabilities.extend(
                [
                    PluginCapability("package_analysis", "1.0", f"{lang} package analysis", 80),
                    PluginCapability(
                        "concurrency_analysis",
                        "1.0",
                        f"{lang} concurrency analysis",
                        75,
                    ),
                ]
            )
        elif lang in ["java", "kotlin", "scala"]:
            capabilities.extend(
                [
                    PluginCapability("jvm_analysis", "1.0", "JVM bytecode analysis", 75),
                    PluginCapability("build_tool_integration", "1.0", "Build tool integration", 70),
                ]
            )

        return capabilities

    @property
    def plugins(self):
        """Get the dictionary of loaded plugins by language."""
        return self._by_lang

    @property
    def supported_languages(self) -> List[str]:
        """Get list of all supported languages (loaded and available)."""
        if self._use_factory:
            return PluginFactory.get_supported_languages()
        else:
            return list(self._by_lang.keys())

    def _get_file_hash(self, content: str) -> str:
        """Compute SHA-256 hash of file content."""
        return hashlib.sha256(content.encode("utf-8")).hexdigest()

    def _should_reindex(self, path: Path, content: str) -> bool:
        """Return True if the file needs to be (re-)indexed."""
        key = str(path)
        with self._file_cache_lock:
            if key not in self._file_cache:
                return True
            cached_mtime, cached_size, cached_hash = self._file_cache[key]
        try:
            stat = path.stat()
            if stat.st_mtime == cached_mtime and stat.st_size == cached_size:
                return False
            return self._get_file_hash(content) != cached_hash
        except OSError:
            return True

    def _match_plugin(self, path: Path) -> IPlugin:
        """Match a plugin for the given file path."""
        # Check explicitly-registered plugins first so mocks and custom plugins take priority
        for p in self._plugins:
            if p.supports(path):
                return p

        # Fall back to lazy loading via factory
        if self._lazy_load and self._use_factory:
            plugin = self._ensure_plugin_for_file(path)
            if plugin:
                return plugin

        # Use advanced routing if available
        if self._enable_advanced and self._router:
            route_result = self._router.get_best_plugin(path)
            if route_result:
                return route_result.plugin

        raise RuntimeError(f"No plugin for {path}")

    def get_plugins_for_file(self, path: Path) -> List[Tuple[IPlugin, float]]:
        """Get all plugins that can handle a file with confidence scores."""
        # Ensure plugin is loaded if using lazy loading
        if self._lazy_load and self._use_factory:
            self._ensure_plugin_for_file(path)

        if self._enable_advanced and self._router:
            route_results = self._router.route_file(path)
            return [(result.plugin, result.confidence) for result in route_results]
        else:
            # Basic fallback
            matching_plugins = []
            for plugin in self._plugins:
                if plugin.supports(path):
                    matching_plugins.append((plugin, 1.0))
            return matching_plugins

    def lookup(self, symbol: str, limit: int = 20) -> SymbolDef | None:
        """Look up symbol definition across all plugins."""
        start_time = time.time()

        try:
            # For symbol lookup, prefer BM25 direct lookup to avoid plugin loading delays
            # Only load plugins if explicitly needed and BM25 fails
            if self._sqlite_store:
                logger.debug("Using BM25 lookup directly for better performance")
                try:
                    import sqlite3

                    conn = sqlite3.connect(self._sqlite_store.db_path)
                    cursor = conn.cursor()

                    # First try symbols table for exact matches
                    cursor.execute(
                        """
                        SELECT s.name, s.kind, s.line_start, s.signature, s.documentation, f.path
                        FROM symbols s
                        JOIN files f ON s.file_id = f.id
                        WHERE s.name = ? OR s.name LIKE ?
                        ORDER BY CASE WHEN s.name = ? THEN 0 ELSE 1 END
                        LIMIT 1
                    """,
                        (symbol, f"%{symbol}%", symbol),
                    )

                    row = cursor.fetchone()
                    if row:
                        name, kind, line, signature, doc, filepath = row
                        conn.close()

                        # Return proper SymbolDef dict
                        return {
                            "symbol": name,
                            "kind": kind,
                            "language": "unknown",  # Not stored in symbols table
                            "signature": signature or f"{kind} {name}",
                            "doc": doc,
                            "defined_in": filepath,
                            "line": line or 1,
                            "span": (0, len(name)),
                        }

                    # Fallback to BM25 if available
                    try:
                        patterns = [
                            f"class {symbol}",
                            f"def {symbol}",
                            f"function {symbol}",
                            symbol,  # Try exact symbol match as fallback
                        ]

                        for pattern in patterns:
                            cursor.execute(
                                """
                                SELECT filepath, snippet(bm25_content, -1, '', '', '...', 20), language
                                FROM bm25_content
                                WHERE bm25_content MATCH ?
                                ORDER BY rank
                                LIMIT 1
                            """,
                                (pattern,),
                            )

                            row = cursor.fetchone()
                            if row:
                                filepath, snippet, language = row

                                # Determine kind from pattern
                                pattern_lower = pattern.lower()
                                if "class" in pattern_lower:
                                    kind = "class"
                                elif "def" in pattern_lower or "function" in pattern_lower:
                                    kind = "function"
                                else:
                                    kind = "symbol"

                                conn.close()

                                return {
                                    "symbol": symbol,
                                    "kind": kind,
                                    "language": language or "unknown",
                                    "signature": snippet,
                                    "doc": None,
                                    "defined_in": filepath,
                                    "line": 1,
                                    "span": (0, len(symbol)),
                                }
                    except sqlite3.OperationalError:
                        # BM25 table doesn't exist, that's fine
                        pass

                    conn.close()
                except Exception as e:
                    logger.error(f"Error in direct symbol lookup: {e}")

            if self._enable_advanced and self._aggregator:
                # Use advanced aggregation
                definitions_by_plugin = {}
                for plugin in self._plugins:
                    try:
                        definition = plugin.getDefinition(symbol)
                        definitions_by_plugin[plugin] = definition
                    except Exception as e:
                        logger.warning(
                            f"Plugin {plugin.lang} failed to get definition for {symbol}: {e}"
                        )
                        definitions_by_plugin[plugin] = None

                result = self._aggregator.aggregate_symbol_definitions(definitions_by_plugin)

                self._operation_stats["lookups"] += 1
                self._operation_stats["total_time"] += time.time() - start_time

                return result
            else:
                # Fallback to basic lookup
                for p in self._plugins:
                    res = p.getDefinition(symbol)
                    if res:
                        self._operation_stats["lookups"] += 1
                        self._operation_stats["total_time"] += time.time() - start_time
                        return res
                return None

        except Exception as e:
            logger.error(f"Error in symbol lookup for {symbol}: {e}", exc_info=True)
            return None

    def _is_document_query(self, query: str) -> bool:
        """Check if the query is looking for documentation.

        Args:
            query: Search query string

        Returns:
            True if this appears to be a documentation query
        """
        query_lower = query.lower()

        # Check against document query patterns
        for pattern in self._compiled_doc_patterns:
            if pattern.search(query_lower):
                return True

        # Check for question words at the beginning
        question_starters = [
            "how",
            "what",
            "where",
            "when",
            "why",
            "can",
            "is",
            "does",
            "should",
        ]
        first_word = query_lower.split()[0] if query_lower.split() else ""
        if first_word in question_starters:
            return True

        return False

    def _expand_document_query(self, query: str) -> List[str]:
        """Expand a document query with related terms for better search coverage.

        Args:
            query: Original search query

        Returns:
            List of expanded query variations
        """
        expanded_queries = [query]  # Always include original
        query_lower = query.lower()

        # Common expansions for documentation queries
        expansions = {
            "install": ["installation", "setup", "getting started", "requirements"],
            "config": [
                "configuration",
                "configure",
                "settings",
                "options",
                "parameters",
            ],
            "api": ["api documentation", "api reference", "endpoint", "method"],
            "how to": ["tutorial", "guide", "example", "usage"],
            "example": ["sample", "snippet", "demo", "code example"],
            "error": ["troubleshoot", "debug", "issue", "problem", "fix"],
            "getting started": ["quickstart", "tutorial", "introduction", "setup"],
            "guide": ["tutorial", "documentation", "walkthrough", "how to"],
            "usage": ["how to use", "example", "api", "reference"],
        }

        # Apply expansions
        for term, related_terms in expansions.items():
            if term in query_lower:
                for related in related_terms:
                    # Replace the term with related term
                    expanded = query_lower.replace(term, related)
                    if expanded != query_lower and expanded not in expanded_queries:
                        expanded_queries.append(expanded)

                # Also add queries with additional terms
                for related in related_terms[:2]:  # Limit to avoid too many queries
                    expanded = f"{query} {related}"
                    if expanded not in expanded_queries:
                        expanded_queries.append(expanded)

        # Add file-specific searches for common documentation files
        if self._is_document_query(query):
            # Extract the main topic from the query
            topic_words = []
            for word in query.lower().split():
                if word not in [
                    "how",
                    "to",
                    "the",
                    "a",
                    "an",
                    "is",
                    "are",
                    "what",
                    "where",
                    "when",
                ]:
                    topic_words.append(word)

            if topic_words:
                topic = " ".join(topic_words[:2])  # Use first two topic words
                expanded_queries.extend(
                    [
                        f"README {topic}",
                        f"{topic} documentation",
                        f"{topic} docs",
                        f"{topic} guide",
                    ]
                )

        # Remove duplicates while preserving order
        seen = set()
        unique_queries = []
        for q in expanded_queries:
            if q not in seen:
                seen.add(q)
                unique_queries.append(q)

        return unique_queries[:10]  # Limit to 10 queries max

    def _is_documentation_file(self, file_path: str) -> bool:
        """Check if a file path is likely a documentation file.

        Args:
            file_path: Path to check

        Returns:
            True if this appears to be a documentation file
        """
        path_lower = file_path.lower()

        for pattern in self._compiled_file_patterns:
            if pattern.search(path_lower):
                return True

        return False

    def _adjust_ranking_for_documents(
        self, query: str, results: List[AggregatedResult]
    ) -> List[AggregatedResult]:
        """Adjust ranking to prioritize documentation files for document queries.

        Args:
            query: Original search query
            results: List of aggregated results

        Returns:
            Re-ranked results with documentation prioritized
        """
        if not self._is_document_query(query):
            return results

        # Separate documentation and code results
        doc_results = []
        code_results = []

        for result in results:
            if self._is_documentation_file(result.primary_result.get("file", "")):
                # Boost documentation files for document queries
                result.rank_score *= 1.5
                result.metadata["doc_boost"] = True
                doc_results.append(result)
            else:
                code_results.append(result)

        # Sort each group by rank score
        doc_results.sort(key=lambda r: r.rank_score, reverse=True)
        code_results.sort(key=lambda r: r.rank_score, reverse=True)

        # Combine with documentation files first
        return doc_results + code_results

    def _get_chunk_content_for_reranking(self, file_path: str) -> str:
        """Return the best chunk content for a file to use as reranker document text.

        Falls back to an empty string if the file has no chunks or the store
        is unavailable — callers must handle the empty case.
        """
        if not self._sqlite_store or not file_path:
            return ""
        try:
            file_id = self._sqlite_store.get_file_id_by_path(file_path)
            if file_id is None:
                return ""
            chunk = self._sqlite_store.find_best_chunk_for_file(file_id, [])
            return chunk.get("content", "") if chunk else ""
        except Exception:
            return ""

    def _symbol_route(self, name: str, kind: Optional[str], limit: int) -> List[Dict]:
        """Query the symbols table directly and return search-result-shaped dicts.

        Prefers definition files over ``__init__.py`` re-exports by sorting
        __init__.py results to the end.  Returns an empty list when the store
        is unavailable or no matching symbol is found (caller falls back to BM25).
        """
        if not self._sqlite_store:
            return []
        try:
            rows = self._sqlite_store.get_symbol(name, kind=kind)
            if not rows and kind:
                # Relax kind constraint and retry
                rows = self._sqlite_store.get_symbol(name, kind=None)
            if not rows:
                return []

            # Sort: non-__init__ files first, then by line number ascending
            def _rank(row: Dict) -> tuple:
                fp = (row.get("file_path") or "").replace("\\", "/")
                is_init = 1 if (fp.endswith("/__init__.py") or fp == "__init__.py") else 0
                return (is_init, row.get("line_start", 0))

            rows.sort(key=_rank)
            results = []
            for row in rows[:limit]:
                file_path = row.get("file_path", "")
                snippet = row.get("signature") or row.get("documentation") or ""
                results.append(
                    {
                        "file": file_path,
                        "line": row.get("line_start", 1),
                        "line_end": row.get("line_end"),
                        "symbol": row.get("name"),
                        "snippet": snippet,
                        "score": 1.0,
                        "language": "unknown",
                    }
                )
            return results
        except Exception as e:
            logger.warning(f"Symbol route failed for '{name}': {e}")
            return []

    def _apply_reranker(
        self,
        query: str,
        candidates: List[Dict],
        limit: int,
        semantic_source: bool = False,
    ) -> List[Dict]:
        """Enrich candidates with chunk text and apply the reranker if configured.

        Returns ``candidates[:limit]`` unchanged when no reranker is set or on error.

        When ``semantic_source=True`` and the configured reranker is text-based
        (flashrank / cross-encoder), skip reranking — text rerankers degrade the
        ordering produced by semantic vector search.

        Pre-filters high-penalty paths (``docs/benchmarks/``, ``htmlcov/``) before
        reranking — those files contain literal query strings and would be
        incorrectly promoted by a text-based reranker.
        """
        if self._reranker is None:
            return candidates[:limit]
        if semantic_source and self._reranker_skips_semantic:
            return candidates[:limit]
        # Exclude paths with penalty >= 1.0 (benchmark/coverage noise files).
        filtered = [c for c in candidates if _path_score_penalty(c.get("file", "")) < 1.0]
        if not filtered:
            filtered = candidates  # nothing passed the filter; fall back to full set
        for c in filtered:
            if not c.get("_rerank_doc"):
                c["_rerank_doc"] = self._get_chunk_content_for_reranking(c.get("file", ""))
        try:
            return self._reranker.rerank(query, filtered, limit)
        except Exception as e:
            logger.warning(f"_apply_reranker failed, returning original order: {e}")
            return candidates[:limit]

    def search(self, query: str, semantic=False, fuzzy=False, limit=20) -> Iterable[SearchResult]:
        """Search for code and documentation across all plugins."""
        start_time = time.time()

        try:
            # Fuzzy (trigram) path for misspelled queries
            if fuzzy and self._sqlite_store:
                logger.info(f"Using fuzzy trigram search for query: {query}")
                try:
                    sym_results = self._sqlite_store.search_symbols_fuzzy(query, limit)
                    file_results = self._sqlite_store.search_files_fuzzy(query, limit)

                    # Merge by file_path, keeping best score
                    merged: Dict[str, Dict] = {}
                    for r in sym_results:
                        fp = r.get("file_path", "")
                        if fp and (fp not in merged or r.get("score", 0) > merged[fp]["score"]):
                            merged[fp] = {
                                "file": fp,
                                "score": r.get("score", 0),
                                "line": r.get("line_start", 1),
                                "snippet": r.get("name", ""),
                                "language": r.get("language", "unknown"),
                            }
                    for r in file_results:
                        fp = r.get("file_path", "")
                        if fp and (fp not in merged or r.get("score", 0) > merged[fp]["score"]):
                            merged[fp] = {
                                "file": fp,
                                "score": r.get("score", 0),
                                "line": 1,
                                "snippet": "",
                                "language": "unknown",
                            }

                    sorted_results = sorted(merged.values(), key=lambda x: x["score"], reverse=True)
                    for item in sorted_results[:limit]:
                        yield item
                    self._operation_stats["searches"] += 1
                    self._operation_stats["total_time"] += time.time() - start_time
                    return
                except Exception as e:
                    logger.warning(f"Fuzzy search failed: {e}")

            # Prefer direct SQLite lexical search for non-semantic queries so the
            # server remains useful even when plugin in-memory indexes are cold.
            if self._sqlite_store and (not semantic or self._semantic_indexer is None):
                # Symbol routing: bypass BM25 for explicit symbol-pattern queries.
                intent, sym_name, kind_hint = classify_query_intent(query)
                if intent == QueryIntent.SYMBOL:
                    sym_results = self._symbol_route(sym_name, kind_hint, limit)
                    if sym_results:
                        logger.info(
                            f"Symbol route hit for '{query}' → '{sym_name}' "
                            f"({len(sym_results)} result(s))"
                        )
                        yield from sym_results
                        self._operation_stats["searches"] += 1
                        self._operation_stats["total_time"] += time.time() - start_time
                        return

                logger.info(f"Using direct lexical search for query: {query}")
                try:
                    tables_to_try = ["bm25_content", "fts_code"]

                    for table in tables_to_try:
                        try:
                            # Fetch an oversampled set so path-based penalties can
                            # surface results that BM25 ranked below the cutoff.
                            fetch_limit = max(limit * 8, 50)
                            results = self._sqlite_store.search_bm25(
                                query, table=table, limit=fetch_limit
                            )
                            if len(results) < fetch_limit:
                                # AND logic found fewer results than the fetch limit —
                                # supplement with OR results so that files matching
                                # some-but-not-all query terms (e.g. a file named
                                # "semantic_preflight.py" lacking the word
                                # "implementation") can enter the candidate pool.
                                or_query = " OR ".join(
                                    t for t in query.split()
                                    if re.match(r"[a-zA-Z0-9_]", t)
                                )
                                if or_query != query:
                                    or_results = self._sqlite_store.search_bm25(
                                        or_query, table=table, limit=fetch_limit
                                    )
                                    and_paths = {
                                        r.get("filepath") or r.get("file_path", "")
                                        for r in results
                                    }
                                    for r in or_results:
                                        fp = r.get("filepath") or r.get("file_path", "")
                                        if fp not in and_paths:
                                            results.append(r)
                            if results:
                                # Apply path-based score penalty and filename token
                                # boost, re-sort, truncate.
                                # FTS5 bm25() scores are negative; adding a positive
                                # penalty degrades rank for non-source paths; a
                                # negative boost from _filename_token_boost improves it.
                                scored = []
                                for result in results:
                                    file_path = result.get("filepath") or result.get(
                                        "file_path", ""
                                    )
                                    raw = result.get("score", 0.0)
                                    adjusted = (
                                        raw
                                        + _path_score_penalty(file_path)
                                        + _filename_token_boost(query, file_path)
                                    )
                                    scored.append((adjusted, file_path, result))
                                scored.sort(key=lambda t: t[0])
                                bm25_candidates = []
                                for adjusted, file_path, result in scored[:limit]:
                                    chunk = None
                                    file_id = result.get("file_id")
                                    if file_id is not None:
                                        try:
                                            chunk = self._sqlite_store.find_best_chunk_for_file(
                                                int(file_id), query.split()
                                            )
                                        except Exception:
                                            chunk = None
                                    bm25_candidates.append(
                                        {
                                            "file": file_path,
                                            "line": (
                                                chunk["line_start"]
                                                if chunk
                                                else result.get("line", 1)
                                            ),
                                            "line_end": chunk["line_end"] if chunk else None,
                                            "symbol": chunk["symbol"] if chunk else None,
                                            "snippet": result.get("snippet", ""),
                                            "score": adjusted,
                                            "language": result.get("language", "unknown"),
                                        }
                                    )
                                bm25_candidates = self._apply_reranker(
                                    query, bm25_candidates, limit
                                )
                                for item in bm25_candidates:
                                    yield {k: v for k, v in item.items() if k != "_rerank_doc"}
                                self._operation_stats["searches"] += 1
                                self._operation_stats["total_time"] += time.time() - start_time
                                return
                        except Exception as e:
                            logger.debug(f"Lexical search in table '{table}' failed: {e}")
                            continue
                except Exception as e:
                    logger.warning(f"Direct lexical search failed: {e}")

            # Semantic queries go directly to the vector index — plugins explicitly
            # return [] for semantic=True, so bypassing them is correct.
            if semantic and self._semantic_indexer:
                logger.info(f"Using semantic indexer for query: {query}")
                try:
                    semantic_results = self._semantic_indexer.search(query=query, limit=limit)
                    candidates = []
                    for result in semantic_results:
                        snippet = result.get("snippet", "")
                        if not snippet and "code" in result:
                            lines = result["code"].split("\n")
                            snippet = "\n".join(lines[:5])
                        file_value = (
                            result.get("relative_path")
                            or result.get("file")
                            or result.get("path")
                            or result.get("file_path")
                            or result.get("filepath")
                            or ""
                        )
                        raw_score = result.get("score", 0.0)
                        # Apply a mild penalty for artifact support paths in semantic results.
                        # The artifacts/ directory holds secondary routing/policy helpers, not
                        # primary implementations, so demote them relative to core code.
                        fp_lower = file_value.replace("\\", "/").lower()
                        if "/artifacts/" in fp_lower:
                            raw_score *= 0.85
                        candidates.append(
                            {
                                "file": file_value,
                                "line": result.get("line", 1),
                                "snippet": snippet,
                                "score": raw_score,
                                "language": result.get("metadata", {}).get("language", "unknown"),
                            }
                        )
                    # Re-sort by adjusted score so penalties take effect before reranking.
                    candidates.sort(key=lambda c: c["score"], reverse=True)
                    candidates = self._apply_reranker(
                        query, candidates, limit, semantic_source=True
                    )
                    for item in candidates:
                        yield {k: v for k, v in item.items() if k != "_rerank_doc"}
                    self._operation_stats["searches"] += 1
                    self._operation_stats["total_time"] += time.time() - start_time
                    return
                except Exception as e:
                    logger.warning(f"Semantic indexer search failed, falling back to plugins: {e}")

            # For search, we may need to search across all languages
            # Load all plugins if using lazy loading
            if self._lazy_load and self._use_factory and len(self._plugins) == 0:
                self._load_all_plugins()

            # If still no plugins, try hybrid or BM25 search directly
            if len(self._plugins) == 0 and self._sqlite_store:
                # Use semantic search if available and requested
                if semantic and self._semantic_indexer:
                    logger.info("No plugins loaded, using semantic search")
                    try:
                        semantic_results = self._semantic_indexer.search(query=query, limit=limit)
                        for result in semantic_results:
                            snippet = result.get("snippet", "")
                            if not snippet and "code" in result:
                                lines = result["code"].split("\n")
                                snippet = "\n".join(lines[:5])
                            file_value = (
                                result.get("relative_path")
                                or result.get("file")
                                or result.get("path")
                                or result.get("file_path")
                                or result.get("filepath")
                                or ""
                            )
                            yield {
                                "file": file_value,
                                "line": result.get("line", 1),
                                "snippet": snippet,
                                "score": result.get("score", 0.0),
                                "language": result.get("metadata", {}).get("language", "unknown"),
                            }
                        self._operation_stats["searches"] += 1
                        self._operation_stats["total_time"] += time.time() - start_time
                        return
                    except Exception as e:
                        logger.error(f"Error in semantic search: {e}")
                        # Fall back to BM25

                # Fall back to BM25-only search
                logger.info("Using BM25 search directly")
                try:
                    import sqlite3

                    conn = sqlite3.connect(self._sqlite_store.db_path)
                    cursor = conn.cursor()

                    # Check if this is a BM25 index
                    cursor.execute(
                        "SELECT name FROM sqlite_master WHERE type='table' AND name='bm25_content'"
                    )
                    if cursor.fetchone():
                        # Use BM25 search
                        cursor.execute(
                            """
                            SELECT 
                                filepath,
                                filename,
                                snippet(bm25_content, -1, '<<', '>>', '...', 20) as snippet,
                                language,
                                rank
                            FROM bm25_content
                            WHERE bm25_content MATCH ?
                            ORDER BY rank
                            LIMIT ?
                        """,
                            (query, limit),
                        )

                        for row in cursor.fetchall():
                            filepath, filename, snippet, language, rank = row
                            yield {
                                "file": filepath,
                                "line": 1,
                                "snippet": snippet,
                                "score": abs(rank),
                                "language": language or "unknown",
                            }

                        conn.close()
                        self._operation_stats["searches"] += 1
                        self._operation_stats["total_time"] += time.time() - start_time
                        return

                    conn.close()
                except Exception as e:
                    logger.error(f"Error in direct BM25 search: {e}")

            # Detect if this is a document query
            is_doc_query = self._is_document_query(query)

            # Expand query if it's a document query
            queries = [query]
            if is_doc_query:
                queries = self._expand_document_query(query)
                logger.info(f"Expanded document query '{query}' to {len(queries)} variations")
                # Force semantic search for natural language queries
                semantic = True

            if self._enable_advanced and self._aggregator:
                # Use advanced aggregation
                all_results_by_plugin = {}
                opts = {
                    "semantic": semantic,
                    "limit": limit * 2 if is_doc_query else limit,
                }

                # Search with all query variations
                for search_query in queries:
                    for plugin in self._plugins:
                        try:
                            results = list(plugin.search(search_query, opts))
                            if results:
                                if plugin not in all_results_by_plugin:
                                    all_results_by_plugin[plugin] = []
                                all_results_by_plugin[plugin].extend(results)
                        except Exception as e:
                            logger.warning(
                                f"Plugin {plugin.lang} failed to search for {search_query}: {e}"
                            )

                # Deduplicate results per plugin (handle both dict results and SearchResult objects)
                for plugin, results in all_results_by_plugin.items():
                    seen = set()
                    unique_results = []
                    for result in results:
                        if hasattr(result, "path"):
                            key = f"{result.path}:{getattr(result, 'line', 0)}"
                        else:
                            key = f"{result.get('file', result.get('path', ''))}:{result.get('line', 0)}"
                        if key not in seen:
                            seen.add(key)
                            unique_results.append(result)
                    all_results_by_plugin[plugin] = unique_results

                # Determine if we have plugin SearchResult dicts (have "path"/"name", not "file")
                # vs BM25 dict results (have "file" key). Plugin results bypass the dict-only aggregator.
                first_result = next(
                    (r for results in all_results_by_plugin.values() for r in results), None
                )
                is_search_result_objects = first_result is not None and (
                    hasattr(first_result, "path")  # dataclass
                    or (
                        isinstance(first_result, dict)
                        and "path" in first_result
                        and "file" not in first_result
                    )
                )

                if is_search_result_objects:
                    # Plugin SearchResult objects: flatten and sort by score, bypassing dict-only aggregator
                    all_search_results = []
                    for results in all_results_by_plugin.values():
                        all_search_results.extend(results)
                    all_search_results.sort(key=lambda r: getattr(r, "score", 0), reverse=True)
                    self._operation_stats["searches"] += 1
                    self._operation_stats["total_time"] += time.time() - start_time
                    for r in all_search_results[:limit]:
                        yield r
                else:
                    # BM25/dict results: use full aggregation pipeline
                    # Configure aggregator for document queries
                    if is_doc_query and self._enable_advanced:
                        doc_criteria = RankingCriteria(
                            relevance_weight=0.5,
                            confidence_weight=0.2,
                            frequency_weight=0.2,
                            recency_weight=0.1,
                            prefer_exact_matches=False,
                            boost_multiple_sources=True,
                            boost_common_extensions=True,
                        )
                        self._aggregator.configure(ranking_criteria=doc_criteria)

                    aggregated_results, stats = self._aggregator.aggregate_search_results(
                        all_results_by_plugin, limit=limit * 2 if is_doc_query else limit
                    )

                    # Adjust ranking for document queries
                    if is_doc_query:
                        aggregated_results = self._adjust_ranking_for_documents(
                            query, aggregated_results
                        )

                    # Apply final limit
                    if limit and len(aggregated_results) > limit:
                        aggregated_results = aggregated_results[:limit]

                    logger.debug(
                        f"Search aggregation stats: {stats.total_results} total, "
                        f"{stats.unique_results} unique, {stats.plugins_used} plugins used, "
                        f"document_query={is_doc_query}"
                    )

                    self._operation_stats["searches"] += 1
                    self._operation_stats["total_time"] += time.time() - start_time

                    # Yield primary results from aggregated results
                    for aggregated in aggregated_results:
                        yield aggregated.primary_result
            else:
                # Fallback to basic search
                # Detect if this is a document query
                is_doc_query = self._is_document_query(query)

                # Expand query if it's a document query
                queries = [query]
                if is_doc_query:
                    queries = self._expand_document_query(query)
                    semantic = True  # Force semantic search for natural language

                opts = {"semantic": semantic, "limit": limit}
                all_results = []

                # Search with all query variations
                for search_query in queries:
                    for p in self._plugins:
                        try:
                            for result in p.search(search_query, opts):
                                all_results.append(result)
                        except Exception as e:
                            logger.warning(
                                f"Plugin {p.lang} failed to search for {search_query}: {e}"
                            )

                # Deduplicate results
                seen = set()
                unique_results = []
                for result in all_results:
                    key = f"{result['file']}:{result['line']}"
                    if key not in seen:
                        seen.add(key)
                        unique_results.append(result)

                # Sort by score if available
                unique_results.sort(key=lambda r: r.get("score", 0.5) or 0.5, reverse=True)

                # Prioritize documentation files for document queries
                if is_doc_query:
                    doc_results = []
                    code_results = []
                    for result in unique_results:
                        if self._is_documentation_file(result.get("file", "")):
                            doc_results.append(result)
                        else:
                            code_results.append(result)
                    unique_results = doc_results + code_results

                # Apply limit
                count = 0
                for result in unique_results:
                    if limit and count >= limit:
                        break
                    yield result
                    count += 1

                self._operation_stats["searches"] += 1
                self._operation_stats["total_time"] += time.time() - start_time

        except Exception as e:
            logger.error(f"Error in search for {query}: {e}", exc_info=True)

    def index_file(self, path: Path, do_semantic: bool = True) -> None:
        """Index a single file if it has changed."""
        try:
            # Ensure path is absolute to avoid relative/absolute path issues
            path = path.resolve()

            # Find the appropriate plugin
            plugin = self._match_plugin(path)

            # Read file content
            try:
                content = path.read_text(encoding="utf-8")
            except UnicodeDecodeError:
                # Try with different encodings
                try:
                    content = path.read_text(encoding="latin-1")
                except Exception as e:
                    logger.error(f"Failed to read {path}: {e}")
                    return

            # Skip if file hasn't changed since last index
            if not self._should_reindex(path, content):
                logger.debug(f"Skipping {path} (unchanged)")
                return

            # Index the file
            start_time = time.time()
            logger.info(f"Indexing {path} with {plugin.lang} plugin")
            shard = plugin.indexFile(path, content)

            # Update file cache after successful indexing
            try:
                stat = path.stat()
                with self._file_cache_lock:
                    self._file_cache[str(path)] = (
                        stat.st_mtime,
                        stat.st_size,
                        self._get_file_hash(content),
                    )
            except OSError:
                pass

            # Record performance if advanced features enabled
            if self._enable_advanced and self._router:
                execution_time = time.time() - start_time
                self._router.record_performance(plugin, execution_time)

            self._operation_stats["indexings"] += 1
            self._operation_stats["total_time"] += time.time() - start_time

            logger.info(
                f"Successfully indexed {path}: {len(shard.get('symbols', []))} symbols found"
            )

            # Semantic indexing for incremental single-file updates
            if do_semantic and self._semantic_indexer:
                try:
                    self._semantic_indexer.index_file(path)
                except Exception as e:
                    logger.warning(f"Semantic indexing failed for {path}: {e}")

        except RuntimeError as e:
            # No plugin found for this file type
            logger.debug(f"No plugin for {path}: {e}")
        except Exception as e:
            logger.error(f"Error indexing {path}: {e}", exc_info=True)

    def get_statistics(self) -> dict:
        """Get statistics about indexed files and languages."""
        try:
            by_language: Dict[str, int] = {}
            with self._file_cache_lock:
                cached_paths = list(self._file_cache.keys())
            for file_path in cached_paths:
                for lang, plugin in self._by_lang.items():
                    if plugin.supports(Path(file_path)):
                        by_language[lang] = by_language.get(lang, 0) + 1
                        break
            return {
                "total": len(cached_paths),
                "by_language": by_language,
            }
        except Exception:
            return {"total": 0, "by_language": {}}

    def index_directory(self, directory: Path, recursive: bool = True) -> Dict[str, int]:
        """
        Index all files in a directory, respecting ignore patterns.

        Args:
            directory: Directory to index
            recursive: Whether to index subdirectories

        Returns:
            Statistics about indexed files
        """
        logger.info(f"Indexing directory: {directory} (recursive={recursive})")

        # Note: We don't use ignore patterns during indexing
        # ALL files are indexed for local search capability
        # Filtering happens only during export/sharing

        # Get all supported extensions
        supported_extensions = get_all_extensions()

        stats = {
            "total_files": 0,
            "indexed_files": 0,
            "ignored_files": 0,
            "failed_files": 0,
            "by_language": {},
        }

        # Walk directory while pruning excluded directories early.
        if recursive:

            def iter_files() -> Iterable[Path]:
                for current_root, dirnames, filenames in os.walk(directory):
                    dirnames[:] = [d for d in dirnames if d not in _INDEX_EXCLUDED_DIRS]
                    root_path = Path(current_root)
                    for filename in filenames:
                        yield root_path / filename

        else:

            def iter_files() -> Iterable[Path]:
                for path in directory.glob("*"):
                    if path.is_file():
                        yield path

        # Collect paths that were successfully indexed for batch semantic embedding
        semantically_indexed_paths: List[Path] = []

        # Instantiate once per index_directory call to avoid stale root between reindex calls
        ignore_mgr = IgnorePatternManager(directory)

        for path in iter_files():
            if not path.is_file():
                continue

            stats["total_files"] += 1

            relative_parts = (
                path.relative_to(directory).parts if path.is_relative_to(directory) else path.parts
            )
            if any(
                part in _INDEX_EXCLUDED_DIRS or part.endswith(".egg-info")
                for part in relative_parts
            ):
                stats["ignored_files"] += 1
                continue

            if path.name in _INDEX_EXCLUDED_FILENAMES:
                stats["ignored_files"] += 1
                continue

            if path.suffix.lower() in _INDEX_EXCLUDED_SUFFIXES:
                stats["ignored_files"] += 1
                continue

            if ignore_mgr.should_ignore(path):
                stats["ignored_files"] += 1
                continue

            # Try to find a plugin that supports this file
            # This allows us to index ALL files, including .env, .key, etc.
            try:
                # First try to match by extension
                if path.suffix in supported_extensions:
                    # skip_semantic=True — we'll batch semantic embed after the loop
                    self.index_file(path, do_semantic=False)
                    stats["indexed_files"] += 1
                    semantically_indexed_paths.append(path.resolve())
                # For files without recognized extensions, try each plugin's supports() method
                # This allows plugins to match by filename patterns (e.g., .env, Dockerfile)
                else:
                    matched = False
                    for plugin in self._plugins:
                        if plugin.supports(path):
                            self.index_file(path, do_semantic=False)
                            stats["indexed_files"] += 1
                            semantically_indexed_paths.append(path.resolve())
                            matched = True
                            break

                    # If no plugin matched but we want to index everything,
                    # we could add a fallback here to index as plaintext
                    # For now, we'll skip unmatched files
                    if not matched:
                        logger.debug(f"No plugin found for {path}")

                # Track by language
                language = get_language_by_extension(path.suffix)
                if language:
                    stats["by_language"][language] = stats["by_language"].get(language, 0) + 1

            except Exception as e:
                logger.error(f"Failed to index {path}: {e}")
                stats["failed_files"] += 1

        # Batch semantic embedding — O(n/1000) API calls instead of O(n)
        stats["semantic_paths_queued"] = len(semantically_indexed_paths)
        stats["semantic_indexer_present"] = self._semantic_indexer is not None
        if self._semantic_indexer and semantically_indexed_paths:
            logger.info(
                f"Batch semantic indexing {len(semantically_indexed_paths)} files "
                f"(embed_batch_size=1000)"
            )
            try:
                sem_stats = self._semantic_indexer.index_files_batch(
                    semantically_indexed_paths, embed_batch_size=1000
                )
                stats["semantic_indexed"] = sem_stats.get("files_indexed", 0)
                stats["semantic_failed"] = sem_stats.get("files_failed", 0)
                stats["semantic_skipped"] = sem_stats.get("files_skipped", 0)
                stats["total_embedding_units"] = sem_stats.get("total_embedding_units", 0)
                logger.info(
                    f"Semantic batch complete: {sem_stats.get('files_indexed', 0)} indexed, "
                    f"{sem_stats.get('files_skipped', 0)} skipped, "
                    f"{sem_stats.get('total_embedding_units', 0)} embedding units"
                )
            except Exception as e:
                logger.error(f"Batch semantic indexing failed: {e}", exc_info=True)
                stats["semantic_error"] = str(e)

        logger.info(
            f"Directory indexing complete: {stats['indexed_files']} indexed, "
            f"{stats['ignored_files']} ignored, {stats['failed_files']} failed"
        )

        return stats

    def search_documentation(
        self, topic: str, doc_types: Optional[List[str]] = None, limit: int = 20
    ) -> Iterable[SearchResult]:
        """Search specifically across documentation files.

        Args:
            topic: Topic to search for (e.g., "installation", "configuration")
            doc_types: Optional list of document types to search (e.g., ["readme", "guide", "api"])
            limit: Maximum number of results

        Returns:
            Search results from documentation files
        """
        # Default document types if not specified
        if doc_types is None:
            doc_types = [
                "readme",
                "documentation",
                "guide",
                "tutorial",
                "api",
                "changelog",
                "contributing",
            ]

        # Build search queries for different document types
        queries = []
        for doc_type in doc_types:
            queries.extend([f"{doc_type} {topic}", f"{topic} {doc_type}", f"{topic} in {doc_type}"])

        # Also search for the topic in common doc filenames
        queries.extend(
            [
                f"README {topic}",
                f"CONTRIBUTING {topic}",
                f"docs {topic}",
                f"documentation {topic}",
            ]
        )

        # Deduplicate queries
        queries = list(dict.fromkeys(queries))

        logger.info(f"Cross-document search for '{topic}' with {len(queries)} query variations")

        # Use the enhanced search with document-specific handling
        all_results = []
        seen = set()

        for query in queries[:10]:  # Limit to 10 queries to avoid too many searches
            for result in self.search(query, semantic=True, limit=limit):
                # Only include documentation files
                if self._is_documentation_file(result.get("file", "")):
                    key = f"{result['file']}:{result['line']}"
                    if key not in seen:
                        seen.add(key)
                        all_results.append(result)

        # Sort by relevance (score) and return top results
        all_results.sort(key=lambda r: r.get("score", 0.5) or 0.5, reverse=True)

        count = 0
        for result in all_results:
            if count >= limit:
                break
            yield result
            count += 1

    def health_check(self) -> Dict[str, Any]:
        """Perform a health check on all components."""
        health = {
            "status": "healthy",
            "components": {
                "dispatcher": {
                    "status": "healthy",
                    "plugins_loaded": len(self._plugins),
                    "languages_supported": len(self.supported_languages),
                    "factory_enabled": self._use_factory,
                    "lazy_loading": self._lazy_load,
                }
            },
            "plugins": {},
            "errors": [],
        }

        # Check plugin health
        for lang, plugin in self._by_lang.items():
            try:
                plugin_health = {
                    "status": "healthy",
                    "class": plugin.__class__.__name__,
                    "semantic_enabled": getattr(plugin, "_enable_semantic", False),
                }
                if hasattr(plugin, "get_indexed_count"):
                    plugin_health["indexed_files"] = plugin.get_indexed_count()
            except Exception as e:
                plugin_health = {"status": "error", "error": str(e)}
                health["errors"].append(f"Plugin {lang}: {str(e)}")

            health["plugins"][lang] = plugin_health

        # Determine overall health
        if len(health["errors"]) > 0:
            health["status"] = "degraded" if len(health["errors"]) < 3 else "unhealthy"

        return health

    def remove_file(self, path: Union[Path, str]) -> None:
        """Remove a file from all indexes.

        Args:
            path: File path to remove
        """
        path = Path(path).resolve()
        logger.info(f"Removing file from index: {path}")

        # Evict from skip cache so a subsequent index_file() call is not skipped
        with self._file_cache_lock:
            self._file_cache.pop(str(path), None)

        try:
            # Remove from SQLite if available
            if self._sqlite_store:
                from ..core.path_resolver import PathResolver

                path_resolver = PathResolver()
                try:
                    relative_path = path_resolver.normalize_path(path)
                    # Get repository ID - for now assume 1
                    # TODO: Properly detect repository
                    self._sqlite_store.remove_file(relative_path, repository_id=1)
                except Exception as e:
                    logger.error(f"Error removing from SQLite: {e}")

            # Remove from semantic index if available
            try:
                plugin = self._match_plugin(path)
                if plugin and hasattr(plugin, "_indexer") and plugin._indexer:
                    plugin._indexer.remove_file(path)
                    logger.info(f"Removed from semantic index: {path}")
            except Exception as e:
                logger.warning(f"Error removing from semantic index: {e}")

            # Update statistics
            self._operation_stats["deletions"] = self._operation_stats.get("deletions", 0) + 1

        except Exception as e:
            logger.error(f"Error removing file {path}: {e}", exc_info=True)

    def move_file(
        self,
        old_path: Union[Path, str],
        new_path: Union[Path, str],
        content_hash: Optional[str] = None,
    ) -> None:
        """Move a file in all indexes.

        Args:
            old_path: Original file path
            new_path: New file path
            content_hash: Optional content hash to verify unchanged content
        """
        old_path = Path(old_path).resolve()
        new_path = Path(new_path).resolve()
        logger.info(f"Moving file in index: {old_path} -> {new_path}")

        try:
            # Move in SQLite if available
            if self._sqlite_store:
                from ..core.path_resolver import PathResolver

                path_resolver = PathResolver()
                try:
                    old_relative = path_resolver.normalize_path(old_path)
                    new_relative = path_resolver.normalize_path(new_path)
                    # Get repository ID - for now assume 1
                    # TODO: Properly detect repository
                    self._sqlite_store.move_file(
                        old_relative,
                        new_relative,
                        repository_id=1,
                        content_hash=content_hash,
                    )
                except Exception as e:
                    logger.error(f"Error moving in SQLite: {e}")

            # Move in semantic index if available
            try:
                plugin = self._match_plugin(new_path)
                if plugin and hasattr(plugin, "_indexer") and plugin._indexer:
                    plugin._indexer.move_file(old_path, new_path, content_hash)
                    logger.info(f"Moved in semantic index: {old_path} -> {new_path}")
            except Exception as e:
                logger.warning(f"Error moving in semantic index: {e}")

            # Update statistics
            self._operation_stats["moves"] = self._operation_stats.get("moves", 0) + 1

        except Exception as e:
            logger.error(f"Error moving file {old_path} -> {new_path}: {e}", exc_info=True)

    async def cross_repo_symbol_search(
        self,
        symbol: str,
        repositories: Optional[List[str]] = None,
        languages: Optional[List[str]] = None,
        max_repositories: int = 10,
    ) -> Dict[str, Any]:
        """
        Search for a symbol across multiple repositories.

        Args:
            symbol: Symbol name to search for
            repositories: Optional list of specific repository IDs
            languages: Optional list of languages to filter by
            max_repositories: Maximum number of repositories to search

        Returns:
            Dictionary containing aggregated search results
        """
        if not self._cross_repo_coordinator:
            raise RuntimeError(
                "Cross-repository search not enabled. Set MCP_ENABLE_MULTI_REPO=true"
            )

        scope = SearchScope(
            repositories=repositories,
            languages=languages,
            max_repositories=max_repositories,
            priority_order=True,
        )

        try:
            result = await self._cross_repo_coordinator.search_symbol(symbol, scope)

            # Convert to dictionary format for MCP tools
            return {
                "query": result.query,
                "total_results": result.total_results,
                "repositories_searched": result.repositories_searched,
                "search_time": result.search_time,
                "results": result.results,
                "repository_stats": result.repository_stats,
                "deduplication_stats": result.deduplication_stats,
            }
        except Exception as e:
            logger.error(f"Cross-repository symbol search failed: {e}")
            return {
                "query": symbol,
                "total_results": 0,
                "repositories_searched": 0,
                "search_time": 0.0,
                "results": [],
                "repository_stats": {},
                "deduplication_stats": {},
                "error": str(e),
            }

    async def cross_repo_code_search(
        self,
        query: str,
        repositories: Optional[List[str]] = None,
        languages: Optional[List[str]] = None,
        file_types: Optional[List[str]] = None,
        semantic: bool = False,
        limit: int = 50,
        max_repositories: int = 10,
    ) -> Dict[str, Any]:
        """
        Search for code patterns across multiple repositories.

        Args:
            query: Search query/pattern
            repositories: Optional list of specific repository IDs
            languages: Optional list of languages to filter by
            file_types: Optional list of file extensions to filter by
            semantic: Whether to use semantic search
            limit: Maximum number of results to return
            max_repositories: Maximum number of repositories to search

        Returns:
            Dictionary containing aggregated search results
        """
        if not self._cross_repo_coordinator:
            raise RuntimeError(
                "Cross-repository search not enabled. Set MCP_ENABLE_MULTI_REPO=true"
            )

        scope = SearchScope(
            repositories=repositories,
            languages=languages,
            file_types=file_types,
            max_repositories=max_repositories,
            priority_order=True,
        )

        try:
            result = await self._cross_repo_coordinator.search_code(query, scope, semantic, limit)

            # Convert to dictionary format for MCP tools
            return {
                "query": result.query,
                "total_results": result.total_results,
                "repositories_searched": result.repositories_searched,
                "search_time": result.search_time,
                "results": result.results,
                "repository_stats": result.repository_stats,
                "deduplication_stats": result.deduplication_stats,
            }
        except Exception as e:
            logger.error(f"Cross-repository code search failed: {e}")
            return {
                "query": query,
                "total_results": 0,
                "repositories_searched": 0,
                "search_time": 0.0,
                "results": [],
                "repository_stats": {},
                "deduplication_stats": {},
                "error": str(e),
            }

    async def get_cross_repo_statistics(self) -> Dict[str, Any]:
        """
        Get statistics about cross-repository search capabilities.

        Returns:
            Dictionary containing repository statistics
        """
        if not self._cross_repo_coordinator:
            return {
                "enabled": False,
                "message": "Cross-repository search not enabled. Set MCP_ENABLE_MULTI_REPO=true",
            }

        try:
            stats = await self._cross_repo_coordinator.get_search_statistics()
            stats["enabled"] = True
            return stats
        except Exception as e:
            logger.error(f"Failed to get cross-repository statistics: {e}")
            return {
                "enabled": True,
                "error": str(e),
                "total_repositories": 0,
                "total_files": 0,
                "total_symbols": 0,
                "languages": [],
                "repository_details": [],
            }

    def _ensure_graph_initialized(self, file_paths: Optional[List[str]] = None) -> bool:
        """
        Ensure graph components are initialized.

        Args:
            file_paths: Optional list of files to build graph from

        Returns:
            True if graph is initialized, False otherwise
        """
        if not CHUNKER_AVAILABLE:
            logger.warning("Graph features not available: TreeSitter Chunker not installed")
            return False

        # If already initialized and no new files, return
        if self._graph_analyzer is not None and file_paths is None:
            return True

        try:
            # Initialize graph builder
            if self._graph_builder is None:
                self._graph_builder = XRefAdapter()

            # Build graph from files
            if file_paths:
                nodes, edges = self._graph_builder.build_graph(file_paths)
                self._graph_nodes = nodes
                self._graph_edges = edges

                # Initialize analyzer and selector
                self._graph_analyzer = GraphAnalyzer(nodes, edges)
                self._context_selector = ContextSelector(nodes, edges)

                logger.info(f"Graph initialized: {len(nodes)} nodes, {len(edges)} edges")
                return True
            else:
                # No files provided and not initialized
                return False

        except Exception as e:
            logger.error(f"Failed to initialize graph: {e}", exc_info=True)
            return False

    def graph_search(
        self,
        query: str,
        expansion_radius: int = 1,
        max_context_nodes: int = 50,
        semantic: bool = False,
        limit: int = 20,
    ) -> Iterable[SearchResult]:
        """
        Search with graph-based context expansion.

        Args:
            query: Search query
            expansion_radius: How far to expand from search results
            max_context_nodes: Maximum context nodes to add
            semantic: Use semantic search
            limit: Maximum search results

        Returns:
            Search results with expanded context
        """
        # First, perform regular search
        search_results = list(self.search(query, semantic=semantic, limit=limit))

        if not search_results:
            return

        # Try to expand with graph context
        if self._context_selector:
            try:
                context_nodes = self._context_selector.expand_search_results(
                    search_results, expansion_radius, max_context_nodes
                )

                # Add context nodes as additional results
                for node in context_nodes:
                    # Check if already in results
                    already_included = any(r.get("file") == node.file_path for r in search_results)
                    if not already_included:
                        yield {
                            "file": node.file_path,
                            "line": node.line_start or 1,
                            "snippet": f"Context: {node.symbol or node.kind}",
                            "score": node.score,
                            "language": node.language,
                            "context": True,
                        }
            except Exception as e:
                logger.error(f"Error expanding search with graph: {e}")

        # Yield original results
        for result in search_results:
            yield result

    def get_context_for_symbols(
        self,
        symbols: List[str],
        radius: int = 2,
        budget: int = 200,
        weights: Optional[Dict[str, float]] = None,
    ) -> Optional[GraphCutResult]:
        """
        Get optimal context for a list of symbols using graph cut.

        Args:
            symbols: Symbol names to find context for
            radius: Maximum distance from symbols
            budget: Maximum number of nodes in context
            weights: Scoring weights

        Returns:
            GraphCutResult or None if graph not available
        """
        if not self._context_selector:
            logger.warning("Context selector not initialized")
            return None

        try:
            # Find nodes matching symbols
            seed_nodes = []
            for node in self._graph_nodes:
                if node.symbol in symbols:
                    seed_nodes.append(node.id)

            if not seed_nodes:
                logger.warning(f"No graph nodes found for symbols: {symbols}")
                return None

            # Select context
            result = self._context_selector.select_context(
                seeds=seed_nodes, radius=radius, budget=budget, weights=weights
            )

            return result

        except Exception as e:
            logger.error(f"Error getting context for symbols: {e}", exc_info=True)
            return None

    def find_symbol_dependencies(self, symbol: str, max_depth: int = 3) -> List[Dict[str, Any]]:
        """
        Find dependencies of a symbol.

        Args:
            symbol: Symbol name
            max_depth: Maximum depth to traverse

        Returns:
            List of dependent symbols with metadata
        """
        if not self._graph_analyzer:
            logger.warning("Graph analyzer not initialized")
            return []

        try:
            # Find node with this symbol
            node_id = None
            for node in self._graph_nodes:
                if node.symbol == symbol:
                    node_id = node.id
                    break

            if not node_id:
                logger.warning(f"Symbol not found in graph: {symbol}")
                return []

            # Get dependencies
            deps = self._graph_analyzer.find_dependencies(node_id, max_depth)

            # Convert to dict format
            return [
                {
                    "symbol": dep.symbol,
                    "file": dep.file_path,
                    "kind": dep.kind,
                    "language": dep.language,
                    "line": dep.line_start,
                }
                for dep in deps
            ]

        except Exception as e:
            logger.error(f"Error finding dependencies for {symbol}: {e}")
            return []

    def find_symbol_dependents(self, symbol: str, max_depth: int = 3) -> List[Dict[str, Any]]:
        """
        Find dependents of a symbol (what depends on it).

        Args:
            symbol: Symbol name
            max_depth: Maximum depth to traverse

        Returns:
            List of dependent symbols with metadata
        """
        if not self._graph_analyzer:
            logger.warning("Graph analyzer not initialized")
            return []

        try:
            # Find node with this symbol
            node_id = None
            for node in self._graph_nodes:
                if node.symbol == symbol:
                    node_id = node.id
                    break

            if not node_id:
                logger.warning(f"Symbol not found in graph: {symbol}")
                return []

            # Get dependents
            dependents = self._graph_analyzer.find_dependents(node_id, max_depth)

            # Convert to dict format
            return [
                {
                    "symbol": dep.symbol,
                    "file": dep.file_path,
                    "kind": dep.kind,
                    "language": dep.language,
                    "line": dep.line_start,
                }
                for dep in dependents
            ]

        except Exception as e:
            logger.error(f"Error finding dependents for {symbol}: {e}")
            return []

    def get_code_hotspots(self, top_n: int = 10) -> List[Dict[str, Any]]:
        """
        Get code hotspots (highly connected nodes).

        Args:
            top_n: Number of hotspots to return

        Returns:
            List of hotspot information
        """
        if not self._graph_analyzer:
            logger.warning("Graph analyzer not initialized")
            return []

        try:
            hotspots = self._graph_analyzer.get_hotspots(top_n)

            return [
                {
                    "symbol": node.symbol,
                    "file": node.file_path,
                    "kind": node.kind,
                    "language": node.language,
                    "line": node.line_start,
                    "score": node.score,
                }
                for node in hotspots
            ]

        except Exception as e:
            logger.error(f"Error getting hotspots: {e}")
            return []



# ----------------------------------------------------------------------------
#
#  Welcome to Baml! To use this generated code, please run the following:
#
#  $ pip install baml
#
# ----------------------------------------------------------------------------

# This file was generated by BAML: please do not edit it. Instead, edit the
# BAML files and re-generate this code using: baml-cli generate
# baml-cli is available with the baml package.

from . import stream_types, types

type_map = {
    "types.ChunkInput": types.ChunkInput,
    "stream_types.ChunkInput": stream_types.ChunkInput,
    "types.ChunkSummary": types.ChunkSummary,
    "stream_types.ChunkSummary": stream_types.ChunkSummary,
    "types.FileSummaryResult": types.FileSummaryResult,
    "stream_types.FileSummaryResult": stream_types.FileSummaryResult,
}



"""Plugin system data models."""

from dataclasses import dataclass, field
from enum import Enum
from pathlib import Path
from typing import Any, Dict, List, Optional, Set

import yaml


class PluginState(Enum):
    """Plugin lifecycle states."""

    DISCOVERED = "discovered"
    LOADED = "loaded"
    INITIALIZED = "initialized"
    STARTED = "started"
    STOPPED = "stopped"
    ERROR = "error"
    DISABLED = "disabled"


class PluginType(Enum):
    """Types of plugins."""

    LANGUAGE = "language"
    INDEXER = "indexer"
    ANALYZER = "analyzer"
    FORMATTER = "formatter"
    CUSTOM = "custom"


@dataclass
class PluginInfo:
    """Information about a plugin."""

    name: str
    version: str
    description: str
    author: str
    plugin_type: PluginType
    language: Optional[str]  # For language plugins
    file_extensions: List[str]  # Supported file extensions
    path: Path  # Path to plugin module
    module_name: str  # Python module name
    class_name: str = "Plugin"  # Name of the plugin class
    dependencies: List[str] = field(default_factory=list)
    config_schema: Dict[str, Any] = field(default_factory=dict)
    metadata: Dict[str, Any] = field(default_factory=dict)

    def __post_init__(self):
        """Validate and normalize data after initialization."""
        if isinstance(self.path, str):
            self.path = Path(self.path)
        if isinstance(self.plugin_type, str):
            self.plugin_type = PluginType(self.plugin_type)


@dataclass
class PluginConfig:
    """Configuration for a plugin."""

    enabled: bool = True
    priority: int = 0  # Higher priority plugins are used first
    settings: Dict[str, Any] = field(default_factory=dict)
    dependencies: List[str] = field(default_factory=list)
    health_check: Dict[str, Any] = field(default_factory=dict)

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "PluginConfig":
        """Create from dictionary."""
        health_check_defaults = {
            "enabled": True,
            "interval_seconds": 60,
            "timeout_seconds": 5,
        }

        health_check = data.get("health_check", {})
        # Merge with defaults
        for key, default_value in health_check_defaults.items():
            if key not in health_check:
                health_check[key] = default_value

        return cls(
            enabled=data.get("enabled", True),
            priority=data.get("priority", 0),
            settings=data.get("settings", {}),
            dependencies=data.get("dependencies", []),
            health_check=health_check,
        )


@dataclass
class PluginInstance:
    """Runtime information about a plugin instance."""

    info: PluginInfo
    config: PluginConfig
    instance: Any  # The actual plugin instance
    state: PluginState = PluginState.DISCOVERED
    error: Optional[str] = None
    load_time: Optional[float] = None  # Time taken to load (seconds)
    last_health_check: Optional[float] = None  # Timestamp of last health check
    health_status: str = "unknown"  # healthy, unhealthy, unknown
    metrics: Dict[str, Any] = field(default_factory=dict)

    @property
    def is_active(self) -> bool:
        """Check if plugin is in an active state."""
        return self.state in (PluginState.INITIALIZED, PluginState.STARTED)

    @property
    def is_error(self) -> bool:
        """Check if plugin is in error state."""
        return self.state == PluginState.ERROR

    @property
    def is_healthy(self) -> bool:
        """Check if plugin is healthy based on last health check."""
        return self.health_status == "healthy"

    def update_metrics(self, metrics: Dict[str, Any]) -> None:
        """Update plugin performance metrics."""
        import time

        self.metrics.update(metrics)
        self.metrics["last_updated"] = time.time()


@dataclass
class PluginLoadResult:
    """Result of plugin loading operation."""

    success: bool
    plugin_name: str
    message: str
    error: Optional[Exception] = None


@dataclass
class PluginSystemConfig:
    """Configuration for the plugin system."""

    plugin_dirs: List[Path] = field(default_factory=list)
    auto_discover: bool = True
    auto_load: bool = True
    validate_interfaces: bool = True
    enable_hot_reload: bool = False
    max_concurrent_loads: int = 5
    load_timeout_seconds: int = 30
    config_file: Optional[Path] = None
    disabled_plugins: Set[str] = field(default_factory=set)
    plugin_configs: Dict[str, PluginConfig] = field(default_factory=dict)

    # Enhanced configuration options
    defaults: Dict[str, Any] = field(default_factory=dict)
    environments: Dict[str, Dict[str, Any]] = field(default_factory=dict)
    loading: Dict[str, Any] = field(default_factory=dict)
    monitoring: Dict[str, Any] = field(default_factory=dict)
    security: Dict[str, Any] = field(default_factory=dict)
    resource_limits: Dict[str, Any] = field(default_factory=dict)
    backup: Dict[str, Any] = field(default_factory=dict)

    def __post_init__(self):
        """Ensure paths are Path objects and set defaults."""
        self.plugin_dirs = [Path(p) if isinstance(p, str) else p for p in self.plugin_dirs]
        if self.config_file and isinstance(self.config_file, str):
            self.config_file = Path(self.config_file)

        if self.config_file and self.config_file.exists():
            self._load_config_file(self.config_file)

        # Set default values for enhanced options
        if not self.defaults:
            self.defaults = {
                "max_file_size": 1048576,
                "cache_enabled": True,
                "timeout_seconds": 10,
                "retry_attempts": 3,
            }

        if not self.loading:
            self.loading = {
                "strategy": "priority",
                "parallel_loading": True,
                "fail_fast": False,
                "retry_failed": True,
            }

        if not self.monitoring:
            self.monitoring = {
                "enable_metrics": True,
                "enable_health_checks": True,
                "metric_collection_interval": 30,
            }

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "PluginSystemConfig":
        """Create from dictionary."""
        config = cls()

        # Basic settings
        if "plugin_dirs" in data:
            config.plugin_dirs = [Path(p) for p in data["plugin_dirs"]]

        config.auto_discover = data.get("auto_discover", True)
        config.auto_load = data.get("auto_load", True)
        config.validate_interfaces = data.get("validate_interfaces", True)
        config.enable_hot_reload = data.get("enable_hot_reload", False)
        config.max_concurrent_loads = data.get("max_concurrent_loads", 5)
        config.load_timeout_seconds = data.get("load_timeout_seconds", 30)

        if "config_file" in data:
            config.config_file = Path(data["config_file"])

        if "disabled_plugins" in data:
            config.disabled_plugins = set(data["disabled_plugins"])

        # Plugin-specific configurations
        if "plugins" in data:
            for name, plugin_data in data["plugins"].items():
                plugin_config = PluginConfig.from_dict(plugin_data)
                # Apply defaults from global config
                if config.defaults:
                    for key, value in config.defaults.items():
                        if key not in plugin_config.settings:
                            plugin_config.settings[key] = value
                config.plugin_configs[name] = plugin_config

        # Enhanced configuration sections
        config.defaults = data.get("defaults", {})
        config.environments = data.get("environments", {})
        config.loading = data.get("loading", {})
        config.monitoring = data.get("monitoring", {})
        config.security = data.get("security", {})
        config.resource_limits = data.get("resource_limits", {})
        config.backup = data.get("backup", {})

        return config

    def get_environment_config(self, environment: Optional[str] = None) -> Dict[str, Any]:
        """Get configuration for specific environment."""
        if not environment:
            import os

            environment = os.getenv("ENVIRONMENT", "development")

        return self.environments.get(environment, {})

    def _load_config_file(self, config_path: Path) -> None:
        """Load config values from YAML without replacing explicit constructor args."""
        try:
            with open(config_path, "r", encoding="utf-8") as handle:
                data = yaml.safe_load(handle) or {}
        except Exception:
            return

        self.auto_discover = data.get("auto_discover", self.auto_discover)
        self.auto_load = data.get("auto_load", self.auto_load)
        self.validate_interfaces = data.get("validate_interfaces", self.validate_interfaces)
        self.enable_hot_reload = data.get("enable_hot_reload", self.enable_hot_reload)
        self.environments = data.get("environments", self.environments)

    def apply_environment_overrides(self, environment: Optional[str] = None) -> None:
        """Apply environment-specific configuration overrides."""
        env_config = self.get_environment_config(environment)

        if "auto_discover" in env_config:
            self.auto_discover = env_config["auto_discover"]
        if "enable_hot_reload" in env_config:
            self.enable_hot_reload = env_config["enable_hot_reload"]
        if "validate_interfaces" in env_config:
            self.validate_interfaces = env_config["validate_interfaces"]
        if "auto_load" in env_config:
            self.auto_load = env_config["auto_load"]


@dataclass
class PluginEvent:
    """Event emitted by the plugin system."""

    event_type: str
    plugin_name: str
    timestamp: float
    data: Dict[str, Any] = field(default_factory=dict)


# Exception classes
class PluginError(Exception):
    """Base exception for plugin system errors."""


class PluginNotFoundError(PluginError):
    """Plugin not found."""


class PluginLoadError(PluginError):
    """Error loading plugin."""


class PluginInitError(PluginError):
    """Error initializing plugin."""


class PluginValidationError(PluginError):
    """Plugin validation failed."""


class PluginConfigError(PluginError):
    """Plugin configuration error."""



"""Go module resolver for handling go.mod dependencies."""

from __future__ import annotations

import logging
import re
from pathlib import Path
from typing import Dict, List, Optional, Tuple

logger = logging.getLogger(__name__)


class GoModuleResolver:
    """Handles Go module resolution and dependency tracking."""

    def __init__(self, root_path: Path):
        self.root_path = root_path
        self.module_cache: Dict[str, GoModule] = {}
        self.go_mod_path = self._find_go_mod()
        self.current_module = self._parse_current_module() if self.go_mod_path else None

    def _find_go_mod(self) -> Optional[Path]:
        """Find the go.mod file in the project."""
        current = self.root_path
        while current != current.parent:
            go_mod = current / "go.mod"
            if go_mod.exists():
                return go_mod
            current = current.parent
        return None

    def _parse_current_module(self) -> Optional["GoModule"]:
        """Parse the current module's go.mod file."""
        if not self.go_mod_path:
            return None

        try:
            content = self.go_mod_path.read_text()
            module = self._parse_go_mod(content)
            if module:
                self.module_cache[module.name] = module
            return module
        except Exception as e:
            logger.error(f"Failed to parse go.mod: {e}")
            return None

    def _parse_go_mod(self, content: str) -> Optional["GoModule"]:
        """Parse go.mod content."""
        lines = content.strip().split("\n")
        module_name = None
        go_version = None
        dependencies: List[ModuleDependency] = []
        replacements: Dict[str, str] = {}

        in_require_block = False
        in_replace_block = False

        for line in lines:
            line = line.strip()

            # Skip comments and empty lines
            if not line or line.startswith("//"):
                continue

            # Module declaration
            if line.startswith("module "):
                module_name = line[7:].strip()
                continue

            # Go version
            if line.startswith("go "):
                go_version = line[3:].strip()
                continue

            # Require block
            if line == "require (":
                in_require_block = True
                continue
            elif line == "replace (":
                in_replace_block = True
                continue
            elif line == ")":
                in_require_block = False
                in_replace_block = False
                continue

            # Parse dependencies
            if in_require_block:
                dep = self._parse_dependency_line(line)
                if dep:
                    dependencies.append(dep)
            elif in_replace_block:
                replacement = self._parse_replacement_line(line)
                if replacement:
                    replacements[replacement[0]] = replacement[1]
            elif line.startswith("require "):
                dep = self._parse_dependency_line(line[8:])
                if dep:
                    dependencies.append(dep)
            elif line.startswith("replace "):
                replacement = self._parse_replacement_line(line[8:])
                if replacement:
                    replacements[replacement[0]] = replacement[1]

        if module_name:
            return GoModule(
                name=module_name,
                version=go_version,
                dependencies=dependencies,
                replacements=replacements,
                path=self.go_mod_path.parent,
            )
        return None

    def _parse_dependency_line(self, line: str) -> Optional["ModuleDependency"]:
        """Parse a dependency line from go.mod."""
        # Remove comments
        if "//" in line:
            line = line[: line.index("//")]
        line = line.strip()

        if not line:
            return None

        # Match module version pairs
        match = re.match(r"(\S+)\s+(\S+)", line)
        if match:
            module_path = match.group(1)
            version = match.group(2)
            return ModuleDependency(module_path, version)
        return None

    def _parse_replacement_line(self, line: str) -> Optional[Tuple[str, str]]:
        """Parse a replacement line from go.mod."""
        # Remove comments
        if "//" in line:
            line = line[: line.index("//")]
        line = line.strip()

        if not line or "=>" not in line:
            return None

        parts = line.split("=>")
        if len(parts) == 2:
            old = parts[0].strip()
            new = parts[1].strip()
            return (old, new)
        return None

    def resolve_import(self, import_path: str) -> Optional[str]:
        """Resolve an import path to its actual location."""
        if not self.current_module:
            return None

        # Check if it's a standard library import
        if self._is_stdlib_import(import_path):
            return f"stdlib:{import_path}"

        # Check if it's an internal import
        if import_path.startswith(self.current_module.name):
            relative_path = import_path[len(self.current_module.name) :].lstrip("/")
            return str(self.current_module.path / relative_path)

        # Check replacements
        for old, new in self.current_module.replacements.items():
            if import_path.startswith(old):
                if new.startswith("./") or new.startswith("../"):
                    # Local replacement
                    replacement_path = self.go_mod_path.parent / new
                    relative_part = import_path[len(old) :].lstrip("/")
                    return str(replacement_path / relative_part)
                else:
                    # Module replacement
                    return import_path.replace(old, new, 1)

        # External dependency
        return f"external:{import_path}"

    def _is_stdlib_import(self, import_path: str) -> bool:
        """Check if an import is from the Go standard library."""
        # Common stdlib packages
        stdlib_prefixes = [
            "archive/",
            "bufio",
            "builtin",
            "bytes",
            "compress/",
            "container/",
            "context",
            "crypto/",
            "database/",
            "debug/",
            "embed",
            "encoding/",
            "errors",
            "expvar",
            "flag",
            "fmt",
            "go/",
            "hash/",
            "html/",
            "image/",
            "index/",
            "io",
            "log/",
            "math/",
            "mime/",
            "net/",
            "os",
            "path/",
            "plugin",
            "reflect",
            "regexp",
            "runtime/",
            "sort",
            "strconv",
            "strings",
            "sync/",
            "syscall",
            "testing/",
            "text/",
            "time",
            "unicode/",
            "unsafe",
        ]

        for prefix in stdlib_prefixes:
            if import_path == prefix.rstrip("/") or import_path.startswith(prefix):
                return True
        return False

    def get_package_files(self, package_path: str) -> List[Path]:
        """Get all Go files in a package directory."""
        if package_path.startswith("stdlib:") or package_path.startswith("external:"):
            return []

        package_dir = Path(package_path)
        if not package_dir.exists() or not package_dir.is_dir():
            return []

        # Get all .go files except test files
        go_files = [f for f in package_dir.glob("*.go") if not f.name.endswith("_test.go")]
        return go_files


class GoModule:
    """Represents a Go module."""

    def __init__(
        self,
        name: str,
        version: Optional[str],
        dependencies: List["ModuleDependency"],
        replacements: Dict[str, str],
        path: Path,
    ):
        self.name = name
        self.version = version
        self.dependencies = dependencies
        self.replacements = replacements
        self.path = path


class ModuleDependency:
    """Represents a module dependency."""

    def __init__(self, module_path: str, version: str):
        self.module_path = module_path
        self.version = version



"""Fail-closed integrity validation for downloaded index artifacts."""

from __future__ import annotations

import hashlib
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Dict, List, Optional

from mcp_server.artifacts.manifest_v2 import ArtifactManifestV2


@dataclass(frozen=True)
class ArtifactIntegrityGateResult:
    """Result payload produced by artifact integrity validation."""

    passed: bool
    reasons: List[str] = field(default_factory=list)
    expected_checksum: Optional[str] = None
    actual_checksum: Optional[str] = None
    manifest_v2_validated: bool = False


def validate_required_metadata_fields(metadata: Dict[str, Any]) -> List[str]:
    """Validate mandatory metadata structure and fields."""
    reasons: List[str] = []

    required_keys = ["checksum", "commit", "branch", "timestamp", "compatibility"]
    for key in required_keys:
        if key not in metadata:
            reasons.append(f"missing key: {key}")

    compatibility = metadata.get("compatibility")
    if compatibility is None:
        return reasons

    if not isinstance(compatibility, dict):
        reasons.append("compatibility must be an object")
        return reasons

    for key in ["schema_version", "embedding_model"]:
        if key not in compatibility:
            reasons.append(f"missing compatibility key: {key}")

    artifact_type = metadata.get("artifact_type", "full")
    if artifact_type not in {"full", "delta"}:
        reasons.append(f"invalid artifact_type: {artifact_type}")
    elif artifact_type == "delta":
        for key in ["base_commit", "target_commit"]:
            if key not in metadata or not metadata.get(key):
                reasons.append(f"missing delta metadata key: {key}")

    return reasons


def _calculate_checksum(file_path: Path) -> str:
    """Calculate SHA256 checksum of a local file."""
    sha256 = hashlib.sha256()
    with open(file_path, "rb") as file_handle:
        for chunk in iter(lambda: file_handle.read(8192), b""):
            sha256.update(chunk)
    return sha256.hexdigest()


def _read_expected_checksum(
    metadata: Dict[str, Any], checksum_path: Optional[Path]
) -> Optional[str]:
    """Resolve expected checksum from sidecar file first, then metadata."""
    if checksum_path and checksum_path.exists():
        contents = checksum_path.read_text().strip()
        if not contents:
            return None
        return contents.split()[0]
    checksum = metadata.get("checksum")
    if checksum is None:
        return None
    return str(checksum)


def _extract_manifest_v2_payload(metadata: Dict[str, Any]) -> Optional[Any]:
    """Extract optional manifest v2 payload from known metadata keys."""
    for key in ["manifest_v2", "artifact_manifest_v2"]:
        if key in metadata:
            return metadata[key]
    return None


def validate_artifact_integrity(
    metadata: Dict[str, Any],
    archive_path: Path,
    checksum_path: Optional[Path] = None,
) -> ArtifactIntegrityGateResult:
    """Validate metadata, checksum, and optional manifest v2 payload."""
    reasons = validate_required_metadata_fields(metadata)

    expected_checksum = _read_expected_checksum(metadata, checksum_path)
    actual_checksum: Optional[str] = None
    if not expected_checksum:
        reasons.append("artifact checksum is required but missing")
    else:
        actual_checksum = _calculate_checksum(archive_path)
        if actual_checksum != expected_checksum:
            reasons.append(
                f"checksum mismatch: expected={expected_checksum}, actual={actual_checksum}"
            )

    manifest_v2_validated = False
    manifest_v2_payload = _extract_manifest_v2_payload(metadata)
    if manifest_v2_payload is not None:
        if not isinstance(manifest_v2_payload, dict):
            reasons.append("manifest_v2 must be an object")
        else:
            try:
                ArtifactManifestV2.from_dict(manifest_v2_payload)
                manifest_v2_validated = True
            except (KeyError, TypeError, ValueError) as exc:
                reasons.append(f"invalid manifest_v2: {exc}")

    return ArtifactIntegrityGateResult(
        passed=not reasons,
        reasons=reasons,
        expected_checksum=expected_checksum,
        actual_checksum=actual_checksum,
        manifest_v2_validated=manifest_v2_validated,
    )



"""
Configuration validation for production deployments.
"""

import os
import secrets
from pathlib import Path
from typing import Dict, List, Optional
from urllib.parse import urlparse

from .environment import Environment, get_environment, is_production
from .settings import Settings


class ConfigurationError(Exception):
    """Base configuration error."""


class SecurityConfigurationError(ConfigurationError):
    """Security-specific configuration error."""


def validate_security_config(settings: Settings) -> List[str]:
    """
    Validate security configuration.

    Args:
        settings: Application settings to validate

    Returns:
        List[str]: List of validation warnings/errors

    Raises:
        SecurityConfigurationError: For critical security issues
    """
    errors = []
    warnings = []

    security = settings.security

    # JWT Secret Key validation
    if len(security.jwt_secret_key) < 32:
        errors.append("JWT secret key must be at least 32 characters long")

    if is_production():
        # Production-specific security checks

        # Check for weak JWT secret
        if security.jwt_secret_key in [
            "your-secret-key",
            "secret",
            "admin123!",
            "changeme",
        ]:
            errors.append("JWT secret key appears to be a default/weak value")

        # Check token expiration times
        if security.access_token_expire_minutes > 60:
            warnings.append("Access token expiration > 60 minutes may be too long for production")

        if security.refresh_token_expire_days > 30:
            warnings.append("Refresh token expiration > 30 days may be too long for production")

        # CORS validation
        if "*" in security.cors_allowed_origins:
            errors.append("CORS wildcard (*) not allowed in production")

        # Rate limiting validation
        if not security.rate_limit_enabled:
            warnings.append("Rate limiting is disabled in production")
        elif security.rate_limit_requests > 1000:
            warnings.append("Rate limit may be too high for production (>1000 req/hour)")

        # Admin settings validation
        if security.default_admin_email == "admin@localhost":
            errors.append("Default admin email must be changed from localhost in production")

    # Password policy validation
    if not any(
        [
            security.password_require_uppercase,
            security.password_require_lowercase,
            security.password_require_numbers,
            security.password_require_special,
        ]
    ):
        warnings.append("Password policy is very weak - no character requirements")

    if security.password_min_length < 8:
        warnings.append("Password minimum length < 8 characters is weak")

    # Combine and check for critical errors
    all_issues = errors + warnings

    if errors and is_production():
        raise SecurityConfigurationError(
            f"Critical security configuration errors: {'; '.join(errors)}"
        )

    return all_issues


def validate_database_config(settings: Settings) -> List[str]:
    """
    Validate database configuration.

    Args:
        settings: Application settings to validate

    Returns:
        List[str]: List of validation warnings/errors
    """
    issues = []
    db = settings.database

    # Parse database URL
    try:
        parsed = urlparse(db.url)
    except Exception as e:
        issues.append(f"Invalid database URL: {e}")
        return issues

    # Production database validation
    if is_production():
        if parsed.scheme == "sqlite":
            issues.append("SQLite not recommended for production - use PostgreSQL")

        if parsed.scheme == "postgresql":
            # PostgreSQL production checks
            if not parsed.password or len(parsed.password) < 12:
                issues.append("Database password should be at least 12 characters for production")

            if parsed.hostname in ["localhost", "127.0.0.1"]:
                issues.append("Database host should not be localhost in production")

        # Connection pool validation
        if db.pool_size < 5:
            issues.append("Database pool size < 5 may be too small for production")
        elif db.pool_size > 50:
            issues.append("Database pool size > 50 may be excessive")

    return issues


def validate_cache_config(settings: Settings) -> List[str]:
    """
    Validate cache configuration.

    Args:
        settings: Application settings to validate

    Returns:
        List[str]: List of validation warnings/errors
    """
    issues = []
    cache = settings.cache

    if is_production():
        if not cache.redis_url:
            issues.append("Redis cache recommended for production environments")
        else:
            # Redis URL validation
            try:
                parsed = urlparse(cache.redis_url)
                if parsed.hostname in ["localhost", "127.0.0.1"]:
                    issues.append("Redis host should not be localhost in production")
            except Exception as e:
                issues.append(f"Invalid Redis URL: {e}")

        # TTL validation
        if cache.default_ttl < 300:
            issues.append("Cache TTL < 5 minutes may cause excessive database load")
        elif cache.default_ttl > 86400:
            issues.append("Cache TTL > 24 hours may cause stale data issues")

    return issues


def validate_logging_config(settings: Settings) -> List[str]:
    """
    Validate logging configuration.

    Args:
        settings: Application settings to validate

    Returns:
        List[str]: List of validation warnings/errors
    """
    issues = []
    logging = settings.logging

    if is_production():
        # Production logging checks
        if logging.level == "DEBUG":
            issues.append("DEBUG logging level not recommended for production")

        if not logging.json_format:
            issues.append("JSON logging format recommended for production")

        if logging.log_request_body or logging.log_response_body:
            issues.append("Request/response body logging may expose sensitive data")

        if logging.log_file:
            # Check log file path
            log_path = Path(logging.log_file)
            if not log_path.parent.exists():
                issues.append(f"Log directory does not exist: {log_path.parent}")

            # Check permissions (if file exists)
            if log_path.exists():
                try:
                    with open(log_path, "a"):
                        pass
                except PermissionError:
                    issues.append(f"Cannot write to log file: {log_path}")

    return issues


def validate_environment_variables() -> List[str]:
    """
    Validate critical environment variables are set.

    Returns:
        List[str]: List of missing or invalid environment variables
    """
    issues = []
    env = get_environment()

    # Critical environment variables for production
    if env == Environment.PRODUCTION:
        required_vars = ["DATABASE_URL", "JWT_SECRET_KEY", "DEFAULT_ADMIN_EMAIL"]

        for var in required_vars:
            if not os.getenv(var):
                issues.append(f"Required environment variable not set: {var}")

        # Check for default/insecure values
        insecure_vars = {
            "JWT_SECRET_KEY": ["your-secret-key", "secret", "changeme"],
            "DEFAULT_ADMIN_PASSWORD": ["admin", "password", "admin123!"],
        }

        for var, insecure_values in insecure_vars.items():
            value = os.getenv(var, "")
            if value in insecure_values:
                issues.append(f"Environment variable {var} has insecure default value")

    return issues


def validate_production_config(
    settings: Optional[Settings] = None,
) -> Dict[str, List[str]]:
    """
    Comprehensive production configuration validation.

    Args:
        settings: Settings to validate (defaults to current settings)

    Returns:
        Dict[str, List[str]]: Validation results by category

    Raises:
        SecurityConfigurationError: For critical security issues
    """
    if settings is None:
        from .settings import get_settings

        settings = get_settings()

    results = {
        "security": validate_security_config(settings),
        "database": validate_database_config(settings),
        "cache": validate_cache_config(settings),
        "logging": validate_logging_config(settings),
        "environment": validate_environment_variables(),
    }

    # Count total issues
    total_issues = sum(len(issues) for issues in results.values())

    if total_issues > 0 and is_production():
        # Log summary for production
        print("\n🔍 Configuration Validation Summary:")
        print(f"Total issues found: {total_issues}")

        for category, issues in results.items():
            if issues:
                print(f"\n{category.upper()} Issues ({len(issues)}):")
                for issue in issues:
                    print(f"  - {issue}")

    return results


def generate_secure_defaults() -> Dict[str, str]:
    """
    Generate secure default values for configuration.

    Returns:
        Dict[str, str]: Secure default environment variables
    """
    return {
        "JWT_SECRET_KEY": secrets.token_urlsafe(32),
        "DEFAULT_ADMIN_PASSWORD": secrets.token_urlsafe(16),
        "CACHE_SECRET_KEY": secrets.token_urlsafe(16),
    }


def create_env_template(environment: Environment) -> str:
    """
    Create environment variable template for specific environment.

    Args:
        environment: Target environment

    Returns:
        str: Environment variable template content
    """
    secure_defaults = generate_secure_defaults()

    if environment == Environment.PRODUCTION:
        return f"""# Production Environment Configuration
# DO NOT commit this file with real values!

# Environment
MCP_ENVIRONMENT=production

# Database (Required)
DATABASE_URL=postgresql://mcp_user:secure_password@db-host:5432/mcp_production

# Security (Required)
JWT_SECRET_KEY={secure_defaults['JWT_SECRET_KEY']}
DEFAULT_ADMIN_EMAIL=admin@yourdomain.com
DEFAULT_ADMIN_PASSWORD={secure_defaults['DEFAULT_ADMIN_PASSWORD']}

# Cache (Required for production)
REDIS_URL=redis://redis-host:6379

# Server
HOST=0.0.0.0
PORT=8000
WORKERS=4

# CORS (Specify your domain)
CORS_ALLOWED_ORIGINS=https://yourdomain.com,https://api.yourdomain.com

# Rate Limiting
RATE_LIMIT_REQUESTS=60
RATE_LIMIT_WINDOW=3600

# Logging
LOG_LEVEL=INFO
LOG_FILE=PathUtils.get_log_path()/app.log
LOG_JSON_FORMAT=true

# Monitoring
PROMETHEUS_ENABLED=true
PROMETHEUS_PORT=8001

# Features
DYNAMIC_PLUGIN_LOADING=true
SEMANTIC_SEARCH_ENABLED=false
"""

    elif environment == Environment.STAGING:
        return f"""# Staging Environment Configuration

# Environment
MCP_ENVIRONMENT=staging

# Database
DATABASE_URL=postgresql://mcp_user:staging_password@staging-db:5432/mcp_staging

# Security
JWT_SECRET_KEY={secure_defaults['JWT_SECRET_KEY']}
DEFAULT_ADMIN_EMAIL=admin@staging.yourdomain.com
DEFAULT_ADMIN_PASSWORD={secure_defaults['DEFAULT_ADMIN_PASSWORD']}

# Cache
REDIS_URL=redis://staging-redis:6379

# Server
HOST=0.0.0.0
PORT=8000
WORKERS=2

# CORS
CORS_ALLOWED_ORIGINS=https://staging.yourdomain.com

# Logging
LOG_LEVEL=DEBUG
LOG_JSON_FORMAT=true

# Monitoring
PROMETHEUS_ENABLED=true
"""

    else:  # Development
        return f"""# Development Environment Configuration

# Environment
MCP_ENVIRONMENT=development

# Database (SQLite for development)
DATABASE_URL=sqlite:///./dev_code_index.db

# Security
JWT_SECRET_KEY={secure_defaults['JWT_SECRET_KEY']}
DEFAULT_ADMIN_EMAIL=admin@localhost
DEFAULT_ADMIN_PASSWORD=admin123!

# Cache (Optional Redis)
# REDIS_URL=redis://localhost:6379

# Server
HOST=127.0.0.1
PORT=8000
DEBUG=true

# CORS (Allow all for development)
CORS_ALLOWED_ORIGINS=*

# Logging
LOG_LEVEL=DEBUG
LOG_JSON_FORMAT=false

# Monitoring
PROMETHEUS_ENABLED=true

# Features
DYNAMIC_PLUGIN_LOADING=true
SEMANTIC_SEARCH_ENABLED=true
"""
